/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard;

import java.io.IOException;
import java.util.Comparator;

import org.simantics.databoard.adapter.AdaptException;
import org.simantics.databoard.adapter.Adapter;
import org.simantics.databoard.adapter.AdapterConstructionException;
import org.simantics.databoard.adapter.AdapterRepository;
import org.simantics.databoard.adapter.RuntimeAdaptException;
import org.simantics.databoard.adapter.RuntimeAdapterConstructionException;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.error.BindingConstructionException;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.error.RuntimeBindingConstructionException;
import org.simantics.databoard.bindingscheme.BindingRepository;
import org.simantics.databoard.bindingscheme.BindingScheme;
import org.simantics.databoard.bindingscheme.GenericBindingScheme;
import org.simantics.databoard.bindingscheme.UnboundScheme;
import org.simantics.databoard.reflection.ReflectionBinding;
import org.simantics.databoard.serialization.Serializer;
import org.simantics.databoard.type.DataType;
import org.simantics.databoard.util.DataValueUtil;
import org.simantics.databoard.util.SingleBinding;

/**
 * This ia a facade class for the binding services.
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class Bindings {

	public static final AdapterRepository adapterCache = new AdapterRepository();
	public static final ReflectionBinding reflectionBinding = new ReflectionBinding();
	public static final UnboundScheme UnboundScheme = new UnboundScheme();
	public static final GenericBindingScheme genericBindingScheme = new GenericBindingScheme();
	/** Default Scheme supports java primitives and others */ 
    public static final BindingRepository defaultScheme;	

	
    /**
     * Get a generic binding to a DataType. The result is a binding to 
     * completely mutable java classes. 
     * 
	 * DataType           | Class of the bound instance
	 * ===================|==================
	 * UnionType          | GenericBinding.TaggedObject.class
	 * OptionType         | ValueContainer.class
	 * RecordType         | Object[].class
	 * BooleanType        | MutableBoolean.class
	 * DoubleType         | MutableDouble.class
	 * FloatType          | MutableFloat.class
	 * ByteType           | MutableByte.class
	 * IntegerType        | MutableInt.class
	 * LongType           | MutableLong.class
	 * StringType         | ValueContainer.class
	 * ArrayType          | ArrayList.class
	 * MapType            | TreeMap.class
	 * VariantType        | Variant.class
     *
	 * @param type
	 * @return binding binding to a mutable class 
     */    
    @SuppressWarnings("unchecked")
	public static <T extends Binding> T getBinding(DataType type) {
    	return (T) genericBindingScheme.getBinding(type);
    }
	
	/**
	 * Get a binding to a Java Class. Type details can be modified with annotations.
	 * Please see the package org.simantics.databoard.annotations. 
	 * <p>
	 *  
	 * The whether the result binding is a completely mutable or not depends on the
	 * provided classes. For instance, fields such as Boolean, Integer, Long
	 * are not mutable, instead MutableBoolean, MutableInteger and MutableLong are.
	 * The length of Object[] is not mutable, but length of List<Object> is. <p> 
	 * 
	 * @see ReflectionBinding  
	 * @param clazz
	 * @return binding
	 * @throws BindingConstructionException
	 */
    @SuppressWarnings("unchecked")
	public static <T extends Binding> T getBinding(Class<?> clazz)
    throws BindingConstructionException
    {
		return (T) reflectionBinding.getBinding(clazz);
    }
    
    /**
	 * Read binding and type from a class. DataType details and parameters
	 * are read as annotations placed in the class.  
	 * (See org.simantics.databoard.annotations)
	 * <p>
	 * As an exception, in the subclasses of {@link Throwable}, the fields of
	 * Throwable are omited.
	 * <p>
	 * This method is used when the caller is 100% sure that binding will be 
	 * constructed without exceptions. Such classes
	 * are all primitive types (Double, Integer, etc, arrays, DataType, ...)
	 * 
	 * This method is unchecked if binding construction to the clazz cannot be trusted.
	 * If construction fails, a RuntimeException is thrown.
	 * 
	 * @param clazz
	 * @return binding
	 * @throws RuntimeBindingConstructionException
     */
    public static Binding getBindingUnchecked(Class<?> clazz)
    throws RuntimeBindingConstructionException
    {
		try {
			return reflectionBinding.getBinding(clazz);
		} catch (BindingConstructionException e) {
			throw new RuntimeBindingConstructionException(e);
		}
    }
    
    /**
	 * Unbound binding is a fake binding that is used with {@link Serializer} for
	 * two purposes. The first is for validating data the format of data with
	 * out instantiating Objects, and the second is for skipping over bytes of
	 * unknown sized objects of a stream.
	 * 
	 * @param type
	 * @return binding
	 */
	@SuppressWarnings("unchecked")
	public static <T extends Binding> T getUnboundBinding(DataType type) {
		return (T) UnboundScheme.getBinding(type);
	}

	/** 
	 * Create Binding Scheme that covers one binding 
	 * 
	 * @param binding 
	 * @return scheme that covers the binding
	 */
	public static BindingScheme schemeOver(Binding binding)
	{
		return new SingleBinding(binding);
	}

    /**
     * Create an adapter that adapts two bindings of the same
     * data type.
     * 
     * @param domain binding of the source instance
     * @param range binding of the result instance
     * @return result adapter
     * @throws AdapterConstructionException 
     */
    public static Adapter getAdapter(Binding domain, Binding range)
    throws AdapterConstructionException
    {
    	return adapterCache.getAdapter(domain, range, false, false);
    }
    
    /**
     * Create an adapter that adapts between two bindings of the same
     * data type.
     * 
     * @param domain binding of the source instance
     * @param range binding of the result instance
     * @return result adapter
     * @throws AdapterConstructionException 
     */
    public static Adapter getAdapterUnchecked(Binding domain, Binding range)
    throws RuntimeAdapterConstructionException
    {
    	try {
			return adapterCache.getAdapter(domain, range, false, false);
		} catch (AdapterConstructionException e) {
			throw new RuntimeAdapterConstructionException(e);
		}
    }

    /**
     * Create a type adapter that adapts instances from one DataType to 
     * another. Type Adapter does the following conversions: 
     * 
     *    o Number Types, e.g. long -> double
     *    o Unit Types, e.g. mph -> km/h
     *    o Record Types, for each field of the range, there must be equal in domain 
     *    o Union Types, for each tag type of the domain, there must be equal in range
     * 
     * {@link AdaptException} is thrown at runtime, if number values are 
     * not compatible, e.g. converting value 500 from Long to Byte.
     * Note, there is also a possibility of precision loss, e.g. when 
     * converting from double to int.
     * 
     * @param domain binding of the source instance
     * @param range binding of the result instance
     * @return result adapter
     * @throws AdapterConstructionException 
     */
    public static Adapter getTypeAdapter(Binding domain, Binding range)
    throws AdapterConstructionException
    {
    	return adapterCache.getAdapter(domain, range, true, false);
    }

    /**
     * Create a type adapter that adapts instances from one DataType to 
     * another. Type Adapter does the following conversions: 
     * 
     *    o Number Types, e.g. long -> double
     *    o Unit Types, e.g. mph -> km/h
     *    o Record Types, for each field of the range, there must be equal in domain 
     *    o Union Types, for each tag type of the domain, there must be equal in range
     * 
     * {@link AdaptException} is thrown at runtime, if number values are 
     * not compatible, e.g. converting value 500 from Long to Byte.
     * Note, there is also a possibility of precision loss, e.g. when 
     * converting from double to int.
     * 
     * @param domain binding of the source instance
     * @param range binding of the result instance
     * @return result adapter
     * @throws AdapterConstructionException 
     */
    public static Adapter getTypeAdapterUnchecked(Binding domain, Binding range)
    {
    	try {
			return adapterCache.getAdapter(domain, range, true, false);
		} catch (AdapterConstructionException e) {
			throw new RuntimeAdapterConstructionException(e);
		}
    }
    
    /**
     * Adapt a value of one type to another. 
     * 
     * @param value
     * @param domain
     * @param range
     * @return adapted value
     * @throws AdapterConstructionException
     * @throws AdaptException
     */
    public static Object adapt(Object value, Binding domain, Binding range)
    throws AdaptException
    {
    	try {
			return adapterCache.getAdapter(domain, range, true, false).adapt(value);
		} catch (AdapterConstructionException e) {
			throw new AdaptException(e);
		}
    }
    
    /**
     * Adapt a value of one type to another
     * 
     * @param value
     * @param domain
     * @param range
     * @return adapted value
     * @throws AdapterConstructionException
     * @throws AdaptException
     */
    public static Object adaptUnchecked(Object value, Binding domain, Binding range)
    throws RuntimeAdapterConstructionException, RuntimeAdaptException
    {
    	try {
    		return  adapterCache.getAdapter(domain, range, true, false).adaptUnchecked(value);
		} catch (RuntimeAdapterConstructionException e) {
			throw new RuntimeAdaptException(new AdaptException(e.getCause()));
		} catch (AdapterConstructionException e) {
			throw new RuntimeAdaptException(new AdaptException(e));
		}
    }    
    
    /**
     * Clone a value to another type to another. Bindings that handle immutable values
     * may return the same instance, others will guarantee a complete copy.  
     * 
     * @param value
     * @param domain
     * @param range
     * @return adapted value
     * @throws AdapterConstructionException
     * @throws AdaptException
     */
    public static Object clone(Object value, Binding domain, Binding range)
    throws AdaptException
    {
    	try {
			return adapterCache.getAdapter(domain, range, true, true).adapt(value);
		} catch (AdapterConstructionException e) {
			throw new AdaptException(e);
		}
    }
    

    /**
     * Clone a value of one binding to another. Bindings that handle immutable values
     * may return the same instance, others will guarantee a complete copy.
     * 
     * @param value
     * @param domain
     * @param range
     * @return adapted value
     * @throws AdapterConstructionException
     * @throws AdaptException
     */
    public static Object cloneUnchecked(Object value, Binding domain, Binding range)
    throws RuntimeAdapterConstructionException, RuntimeAdaptException
    {
    	try {
			return adapterCache.getAdapter(domain, range, true, true).adapt(value);
		} catch (AdaptException e) {
			throw new RuntimeAdaptException(e);
		} catch (RuntimeAdapterConstructionException e) {
			throw new RuntimeAdaptException(new AdaptException(e.getCause()));
		} catch (AdapterConstructionException e) {
			throw new RuntimeAdaptException(new AdaptException(e));
		}
    }
    
    /**
     * Compares its two data values for order.  Returns a negative integer,
     * zero, or a positive integer as the first argument is less than, equal
     * to, or greater than the second.<p>
     * 
     * DataTypes of b1 and b2 are not equal then data types are compared. <p>
     *
     * The comparison function is defined at 
     * https://www.simantics.org/wiki/index.php/Org.simantics.datatype_Manual#CompareTo_and_Equals <p>
     * 
     * Note, comparing 2 different number types will not result a value comparison.
     * Instead values have the following type precedence ByteType, IntegerType, LongType,
     * FloatType, and the highest DoubleType. <p>
     *
     * @param b1 Binding of o1 
     * @param o1 the first object to be compared.
     * @param b2 Binding of o2
     * @param o2 the second object to be compared.
     * @return a negative integer, zero, or a positive integer as the
     * 	       first argument is less than, equal to, or greater than the
     *	       second.
     * @throws BindingException if object cannot be handled by a binding
     */
    public static int compare(Binding b1, Object o1, Binding b2, Object o2) 
    throws BindingException
    {
    	return DataValueUtil.compare(b1, o1, b2, o2);
    }
    
    /**
     * Compare two data values for equality. <p>
     * 
     * Note, comparing 2 different number types will not result a value comparison.
     * Instead values have the following type precedence ByteType, IntegerType, LongType,
     * FloatType, and the highest DoubleType. <p>
     * 
     * @param b1
     * @param o1
     * @param b2
     * @param o2
     * @return true if equal
     * @throws BindingException
     */
    public static boolean equals(Binding b1, Object o1, Binding b2, Object o2)
    throws BindingException
    {
    	return DataValueUtil.equals(b1, o1, b2, o2);
    }
    
    public static Comparator<Object> createComparator(final Binding b1, final Binding b2)
    {
    	return DataValueUtil.createComparator(b1, b2);
    }
    
    /**
     * Print the content of an object as a structure.
     * Utility function for debug purposes.
     * 
     * @param o
     * @return content
     */
    public static String toString(Object o) {
    	try {
			Binding b = Bindings.getBinding( o.getClass() );
			return b.printValueDefinition(o, true);
		} catch (BindingConstructionException e) {
			return "<error "+e.getClass().getName()+" "+e.getMessage()+">";
		} catch (IOException e) {
			return "<error "+e.getClass().getName()+" "+e.getMessage()+">";
		} catch (BindingException e) {
			return "<error "+e.getClass().getName()+" "+e.getMessage()+">";
		}
    }
	
    static {
        try {
            defaultScheme = new BindingRepository();
			defaultScheme.add( getBinding( Integer.class ) );
	        defaultScheme.add( getBinding( Byte.class ) );
	        defaultScheme.add( getBinding( Float.class ) );
	        defaultScheme.add( getBinding( Double.class ) );
	        defaultScheme.add( getBinding( Long.class ) );
	        defaultScheme.add( getBinding( String.class ) );
	        defaultScheme.add( getBinding( boolean[].class ) );
	        defaultScheme.add( getBinding( int[].class ) );
	        defaultScheme.add( getBinding( byte[].class ) );
	        defaultScheme.add( getBinding( float[].class ) );
	        defaultScheme.add( getBinding( double[].class ) );
	        defaultScheme.add( getBinding( long[].class ) );
	        defaultScheme.add( getBinding( boolean[].class ) );
		} catch (BindingConstructionException e) {
			// Unexpected
			throw new RuntimeBindingConstructionException(e);
		}
    }
	
}

