/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.accessor.binary;

import java.io.File;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

import org.simantics.databoard.Bindings;
import org.simantics.databoard.accessor.Accessor;
import org.simantics.databoard.accessor.CloseableAccessor;
import org.simantics.databoard.accessor.error.AccessorConstructionException;
import org.simantics.databoard.accessor.error.AccessorException;
import org.simantics.databoard.accessor.event.Event;
import org.simantics.databoard.accessor.event.InvalidatedEvent;
import org.simantics.databoard.accessor.file.FileAccessor;
import org.simantics.databoard.accessor.impl.ListenerEntry;
import org.simantics.databoard.accessor.interestset.InterestSet;
import org.simantics.databoard.accessor.reference.AccessorReference;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.serialization.SerializationException;
import org.simantics.databoard.serialization.Serializer;
import org.simantics.databoard.type.ArrayType;
import org.simantics.databoard.type.BooleanType;
import org.simantics.databoard.type.ByteType;
import org.simantics.databoard.type.DataType;
import org.simantics.databoard.type.DoubleType;
import org.simantics.databoard.type.FloatType;
import org.simantics.databoard.type.IntegerType;
import org.simantics.databoard.type.LongType;
import org.simantics.databoard.type.MapType;
import org.simantics.databoard.type.OptionalType;
import org.simantics.databoard.type.RecordType;
import org.simantics.databoard.type.StringType;
import org.simantics.databoard.type.UnionType;
import org.simantics.databoard.type.VariantType;
import org.simantics.databoard.util.binary.BinaryFile;
import org.simantics.databoard.util.binary.Blob;
import org.simantics.databoard.util.binary.RandomAccessBinary;

/**
 * BinaryObject is an accessor to a binary object, usually a random access file.
 * BinaryObject cannot handle files of recursive types.  
 * <p>
 * The file can be opened once. It may not be modified by any other instance other than
 * accessor while accessors are beign used. You must not create more than one
 * instance of BinaryObjects for a file.
 * 
 * Binary Object does not close the underlying file.
 * 
 * @see BinaryArray
 * @see BinaryBoolean
 * @see BinaryByte
 * @see BinaryDouble
 * @see BinaryFloat
 * @see BinaryInteger
 * @see BinaryLong
 * @see BinaryMap
 * @see BinaryOptional
 * @see BinaryRecord
 * @see BinaryString
 * @see BinaryUnion
 * @see BinaryVariant
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public abstract class BinaryObject implements Accessor, FileAccessor, CloseableAccessor {
    
	/** Hard link to the parent object, <code>null</code> if this is root */
	protected BinaryObject parent;	
	/** Listeners */
	protected ListenerEntry listeners = null;
	/** Random access binary object */
	protected Blob b;
	/** Type */
	protected DataType type;
	/** File, optional */
	protected File file;
		
	BinaryObject(BinaryObject parent, Blob blob, DataType type) {
		this.parent = parent;
		this.b = blob;
		this.type = type;
		
		if (parent!=null) {
			file = parent.file();
		} else {
			RandomAccessBinary sourceBinary = b.getSource();
			if (sourceBinary instanceof BinaryFile) {
				BinaryFile bf = (BinaryFile) sourceBinary;
				file = bf.file();
			}
		}		
	}
	
	public DataType type() {
		return type;
	}
		
	public void flush() throws AccessorException {
		try {
			b.flush();
		} catch (IOException e) {
			throw new AccessorException(e);
		}
	}
	
	public RandomAccessBinary getSource() {
		RandomAccessBinary result = b;
		while (result instanceof Blob) result = ((Blob)result).getParent();
		return result;
	}
	
	public void close() throws AccessorException {
		try {
			if (parent!=null) {
				((FileAccessor) parent).close();
				return;
			}
			
			// Root Object
			if (b==null) return;
			RandomAccessBinary rab = getSource();
			rab.flush();			
			rab.close();
			b = null;
		} catch (IOException e) {
			throw new AccessorException(e);
		}
	}
	
	/**
	 * Get file if the binary object is based on binary file. 
	 * 
	 * @return file or <code>null</code>
	 */
	public File file() {
		return file;
	}
	
	public RandomAccessBinary getBinary() {
		return b;
	}
	
	@Override
	public void addListener(Listener listener, InterestSet interestSet, AccessorReference path) throws AccessorException {
		listeners = ListenerEntry.link(listeners, listener, interestSet, path);
	}
		
	protected ListenerEntry detachListener(Listener listener) throws AccessorException {
		ListenerEntry e = listeners;
		ListenerEntry p = null;
		while (e!=null) {
			// Found match
			if (e.listener == listener) {
				// The match was the first entry of the linked list
				if (p==null) {
					listeners = e.next;
					return e;
				}
				// Some other entry, unlink e
				p.next = e.next;
				return e;
			}
			p = e;
			e = e.next;
		}
		return null;		
	}
	
	@Override
	public void removeListener(Listener listener) throws AccessorException {
		detachListener(listener);
	}	
	
	/**
	 * Write a new value and flush the buffer.
	 * 
	 * @param binding
	 * @param newValue
	 */
	@Override
	public void setValue(Binding binding, Object newValue)
			throws AccessorException {
		try {
			setValueNoflush(binding, newValue);
			b.flush();
		} catch (IOException e) {
			throw new AccessorException(e);
		}
	}
	
	/**
	 * Write a new value and don't flush the buffer
	 * 
	 * @param binding
	 * @param newValue
	 * @throws AccessorException
	 */
	public abstract void setValueNoflush(Binding binding, Object newValue) throws AccessorException;

	@Override
	public Object getValue(Binding binding) throws AccessorException {
		try {
			Serializer s = binding.serializer();
			b.position(0L);
			return s.deserialize(b, null);
		} catch(BindingException e) {		
			throw new AccessorException(e);
		} catch (IOException e) {
			throw new AccessorException(e);
		} catch (SerializationException e) {
			throw new AccessorException(e);
		}
	}
	
	/**
	 * Send notification that this accessor has been detached from the parent
	 */
	void invalidatedNotification() {
		ListenerEntry le = listeners;
		while (le!=null) {			
			InterestSet is = le.getInterestSet();
			if (is.inNotifications()) {
				InvalidatedEvent e = new InvalidatedEvent();
				le.emitEvent(e);
			}
			le = le.next;
		}		
	}	
	
	/**
	 * Apply a change set that has events for the particular accessor. 
	 * There are no sub-accessor events. Does not flush buffer.
	 * 
	 * @param cs
	 * @param makeRollback
	 * @return rollback-event
	 * @throws AccessorException
	 */
	abstract Event applyLocal(Event e, boolean makeRollback) throws AccessorException;
	
	@Override
	public void apply(List<Event> cs, LinkedList<Event> rollback) throws AccessorException {
		try {
			boolean makeRollback = rollback != null;
			for (Event e : cs) {
				// Accessor
				BinaryObject a = e.reference == null ? this : (BinaryObject) getAccessor(e.reference);
				// Apply changes				
				Event rbe = a.applyLocal(e, makeRollback);
				if (makeRollback) {
					rbe.reference = e.reference;
					rollback.addFirst( rbe );
				}
			}
			// Flush after successful transaction
			try {
				b.flush();
			} catch (IOException e1) {
				throw new AccessorException(e1);
			}
		} catch (AccessorConstructionException ae) {
			// Attempt to flush, don't report if fails
			// The contract is that all events that did go thru, created a rollback
			// and those must be flushed
			try {
				b.flush();
			} catch (IOException e1) {
			}
			throw new AccessorException(ae);
		}
		
	}	
	
	@Override
	public String toString() {		
		try {
			DataType type = type();
			Binding binding = Bindings.getBinding(type);
			Object instance = getValue(binding);
			return "Accessor("+binding.printValueDefinition(instance, true)+")";
		} catch (AccessorException e) {
			return "Accessor(error="+e.getMessage()+")";
		} catch (BindingException e) {
			return "Accessor(error="+e.getMessage()+")";
		} catch (IOException e) {
			return "Accessor(error="+e.getMessage()+")";
		}
	}		
	
	public static BinaryObject createAccessor(RandomAccessBinary binary, DataType type) throws AccessorConstructionException
	{
		try {
			Blob blob = binary instanceof Blob ? (Blob) binary : new Blob(binary);

			if (type instanceof BooleanType) {
				return new BinaryBoolean(null, blob, (BooleanType) type);
			}
			if (type instanceof ByteType) {
				return new BinaryByte(null, blob, (ByteType)type);
			}
			if (type instanceof IntegerType) {
				return new BinaryInteger(null, blob, (IntegerType)type);
			}
			if (type instanceof LongType) {
				return new BinaryLong(null, blob, (LongType)type);
			}
			if (type instanceof FloatType) {
				return new BinaryFloat(null, blob, (FloatType)type);
			}
			if (type instanceof DoubleType) {
				return new BinaryDouble(null, blob, (DoubleType)type);
			}
			if (type instanceof StringType) {
				return new BinaryString(null, blob, (StringType)type);
			}			
			if (type instanceof OptionalType) {
				return new BinaryOptional(null, blob, (OptionalType)type);
			}			
			if (type instanceof UnionType) {
				return new BinaryUnion(null, blob, (UnionType)type);
			}			
			if (type instanceof RecordType) {
				return new BinaryRecord(null, blob, (RecordType)type);
			}			
			if (type instanceof VariantType) {
				return new BinaryVariant(null, blob, (VariantType)type);
			}			
			if (type instanceof MapType) {
				return new BinaryMap(null, blob, (MapType)type);
			}			
			if (type instanceof ArrayType) {
				return new BinaryArray(null, blob, (ArrayType)type);
			}						
			throw new AccessorConstructionException("Can not create accessor to "+type);
		} catch (IOException e) {
			throw new AccessorConstructionException(e);
		}
	}

	BinaryObject createSubAccessor(DataType type, long position, long length) 
	throws AccessorConstructionException {
		Blob sb = b.createSubBlob(position, length);		
		
		if (type instanceof BooleanType) {
			return new BinaryBoolean(this, sb, (BooleanType) type);
		}
		if (type instanceof ByteType) {
			return new BinaryByte(this, sb, (ByteType)type);
		}
		if (type instanceof IntegerType) {
			return new BinaryInteger(this, sb, (IntegerType)type);
		}
		if (type instanceof LongType) {
			return new BinaryLong(this, sb, (LongType)type);
		}
		if (type instanceof FloatType) {
			return new BinaryFloat(this, sb, (FloatType)type);
		}
		if (type instanceof DoubleType) {
			return new BinaryDouble(this, sb, (DoubleType)type);
		}
		if (type instanceof StringType) {
			return new BinaryString(this, sb, (StringType)type);
		}
		if (type instanceof OptionalType) {
			return new BinaryOptional(this, sb, (OptionalType)type);
		}
		if (type instanceof UnionType) {
			return new BinaryUnion(this, sb, (UnionType)type);
		}
		if (type instanceof VariantType) {
			return new BinaryVariant(this, sb, (VariantType)type);
		}
		if (type instanceof ArrayType) {
			return new BinaryArray(this, sb, (ArrayType)type);
		}
		if (type instanceof MapType) {
			return new BinaryMap(this, sb, (MapType)type);
		}
		if (type instanceof RecordType) {
			return new BinaryRecord(this, sb, (RecordType)type);
		}
		throw new AccessorConstructionException("Can not create accessor to "+type);
	}	
	
}

