/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.accessor.binary;

import java.io.IOException;
import java.lang.ref.Reference;
import java.lang.ref.SoftReference;
import java.util.ArrayList;
import java.util.List;

import org.simantics.databoard.Bindings;
import org.simantics.databoard.accessor.Accessor;
import org.simantics.databoard.accessor.RecordAccessor;
import org.simantics.databoard.accessor.error.AccessorConstructionException;
import org.simantics.databoard.accessor.error.AccessorException;
import org.simantics.databoard.accessor.error.ReferenceException;
import org.simantics.databoard.accessor.event.Event;
import org.simantics.databoard.accessor.event.ValueAssigned;
import org.simantics.databoard.accessor.file.FileRecordAccessor;
import org.simantics.databoard.accessor.impl.ListenerEntry;
import org.simantics.databoard.accessor.interestset.InterestSet;
import org.simantics.databoard.accessor.interestset.RecordInterestSet;
import org.simantics.databoard.accessor.reference.AccessorReference;
import org.simantics.databoard.accessor.reference.FieldIndexReference;
import org.simantics.databoard.accessor.reference.FieldNameReference;
import org.simantics.databoard.accessor.reference.LabelReference;
import org.simantics.databoard.adapter.AdaptException;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.RecordBinding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.mutable.Variant;
import org.simantics.databoard.serialization.SerializationException;
import org.simantics.databoard.serialization.Serializer;
import org.simantics.databoard.type.RecordType;
import org.simantics.databoard.util.binary.Blob;
import org.simantics.databoard.util.binary.RuntimeSerializerConstructionException;

/**
 * Accessor to a Binary Record.
 * Get and set field operations scan the file. 
 * File operations are most efficient if field-specific subaccessors are acquired
 * and used instead of {@link #setFieldValue(int, Binding, Object)} and
 * {@link #getFieldValue(int, Binding)}.
 * <p>
 * Note, To increase the random access performance of the record, create sub-accessors of
 * its fields. 
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class BinaryRecord extends BinaryObject implements RecordAccessor, FileRecordAccessor {

	/** Accessors to children */
	Reference<BinaryObject>[] children;
	
	RecordBinding binding;
	
	@SuppressWarnings("unchecked")
	public BinaryRecord(BinaryObject parent, Blob blob, RecordType type) throws AccessorConstructionException {
		super(parent, blob, type);
		if (type.isReferable()) throw new AccessorConstructionException("Refereable record are not supported");
		
		binding = Bindings.getBinding(type);		
		int count = type.count();
		children = new Reference[count];
	}

	@Override
	public RecordType type() {
		return (RecordType) type;
	}	
	
	@Override
	public int count() {
		return type().count();
	}
	
	/**
	 * Get start position of a field
	 * 
	 * @param fieldIndex
	 * @return
	 * @throws AccessorException
	 */
	long getStartPosition(int fieldIndex) throws AccessorException {
		try {
			// sa, saIndex = getFloorExistingAccessor 
			BinaryObject sa = getExistingAccessor(fieldIndex);
			if (sa!=null) return sa.b.getStartPositionInSourceBinary();
			int saIndex = -1;
			for (int i=fieldIndex; i>=0; i--) {
				sa = getExistingAccessor(i);
				if (sa!=null) {
					saIndex = i;
					break;
				}
			}
				
			long pos = saIndex==-1 ? 0L : sa.b.getStartPositionInSourceBinary();
			b.position(pos);
			if (saIndex<0) saIndex = 0;
			for (int i=saIndex; i<fieldIndex; i++) {
				Binding cb = binding.getComponentBinding(i);
				Serializer cs = cb.serializer();
				cs.skip(b);
			}
			return b.position();
		} catch (IOException ioe) {
			throw new AccessorException( ioe ); 
		} catch (SerializationException se) {
			throw new AccessorException( se ); 
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T extends Accessor> T getFieldAccessor(String fieldName) throws AccessorConstructionException
	{
		int fieldIndex = type().getComponentIndex(fieldName);
		if (fieldIndex<0) throw new AccessorConstructionException("Field "+fieldName+" does not exist");
		return (T) getFieldAccessor(fieldIndex);
	}	
	
	@SuppressWarnings("unchecked")
	@Override
	public <T extends Accessor> T getFieldAccessor(int index) throws AccessorConstructionException
	{
		if (index<0 || index>=count()) throw new ReferenceException("Field index ("+index+") out of bounds ("+count()+")");
		
		try {
			// Get existing or create new
			Reference<BinaryObject> ref = children[index];
			BinaryObject sa = (ref!=null)?(BinaryObject)ref.get():null;
			
			if (sa==null) {
				// Instantiate new accessor
				Binding cb = binding.getComponentBinding(index);
				Serializer cs = cb.serializer(); 
				long pos = getStartPosition(index);
				b.position(pos);
				cs.skip(b);
				long len = b.position() - pos;
				
				sa = createSubAccessor(cb.type(), pos, len);
				children[index] = new SoftReference<BinaryObject>(sa);

				// Add component interest sets
				ListenerEntry le = listeners;
				while (le!=null) {				
					RecordInterestSet is = le.getInterestSet();
					InterestSet cis = is.getComponentInterest(index); 
					if (cis != null) {
						try {
							AccessorReference childPath = AccessorReference.concatenate( le.path, new FieldIndexReference(index) );
							sa.addListener(le.listener, cis, childPath);
						} catch (AccessorException e) {
							throw new AccessorConstructionException(e);
						}
					}
					le = le.next;
				}				
			}		
			
			return (T) sa;
		} catch (IOException e) {
			throw new AccessorConstructionException(e);
		} catch (AccessorException e) {
			throw new AccessorConstructionException(e);
		} catch (SerializationException e) {
			throw new AccessorConstructionException(e);
		}
	}

	/**
	 * Get existing sub accessor
	 * @param index
	 * @return sub-accessor or <code>null</code>
	 */
	BinaryObject getExistingAccessor(int index)
	{
		if (index<0 || index>=count()) throw new RuntimeException("Field index ("+index+") out of bounds ("+count()+")");
		
		// Get existing or create new
		Reference<BinaryObject> ref = children[index];
		BinaryObject accessor = (ref!=null)?(BinaryObject)ref.get():null;
		return accessor;
	}	
	
	@Override
	public Object getFieldValue(int index, Binding fieldBinding)
			throws AccessorException {
		try {
			b.position( getStartPosition(index) );
			List<Object> ids = new ArrayList<Object>(0);
			return fieldBinding.serializer().deserialize(b, ids);
		} catch (BindingException e) {
			throw new AccessorException(e);
		} catch (IOException e) {
			throw new AccessorException(e);
		} catch (RuntimeSerializerConstructionException e) {
			throw new AccessorException(e);
		} catch (SerializationException e) {
			throw new AccessorException(e);
		}
	}
	
	@Override
	public void setFieldValue(int index, Binding fieldBinding,
			Object value) throws AccessorException {
		setFieldValueNoflush(index, fieldBinding, value);
		flush();
	}	

	@Override
	public void setFieldValueNoflush(int index, Binding cb, Object cv)
	throws AccessorException {
		
		try {			
			
			BinaryObject sa = getExistingAccessor(index);				
			if (sa!=null) {
				// Recursive write using existing sub accessor
				sa.setValueNoflush(cb, cv);
				return;
			}
			
			// Init
			long pos = getStartPosition(index);
			Serializer cs = cb.serializer();
						
			// The size might need adjusting
			if (cs.getConstantSize() == null) {
				// Old size
				b.position( pos );
				cs.skip(b, null);
				long oldLen = b.position() - pos;
				
				// New size
				long newLen = cs.getSize(cv, null);
				if (newLen>oldLen) {
					b.position(pos+oldLen);
					b.insertBytes(newLen - oldLen);
				} else {
					b.position(pos+newLen);
					b.removeBytes(oldLen - newLen);
				}
			}
			
			// Write
			b.position(pos);
			cs.serialize(b, null, cv);
			
			// Notify Listeners
			ListenerEntry le = listeners;
			while (le!=null) {				
				RecordInterestSet is = le.getInterestSet();
				if (is.inNotificationsOf(index)) {					
					Variant newValue = is.inValuesOf(index) ? new Variant(cb, cv) : null;
					if (is.inValuesOf(index)) newValue = new Variant(cb, cb.isImmutable() ? cv : cb.clone(cv));
					
					Event e = new ValueAssigned(new FieldIndexReference(index), newValue);
					le.emitEvent(e);
				}
				le = le.next;
			}				
		} catch (BindingException e) {
			throw new AccessorException(e);
		} catch (SerializationException e) {
			throw new AccessorException(e);
		} catch (IOException e) {
			throw new AccessorException(e);
		} catch (AdaptException e) {
			throw new AccessorException(e);
		}		
	}
		
	@Override
	public void addListener(Listener listener, InterestSet interestSet,
			AccessorReference path) throws AccessorException {
		RecordInterestSet is = (RecordInterestSet) interestSet;
		super.addListener(listener, interestSet, path);
		// Apply component interest set to existing sub-accessors
		if (is.componentInterests!=null) {
			for (int i=0; i<count(); i++) {
				InterestSet cis = is.getComponentInterest(i);
				if (cis==null) continue;
				BinaryObject sa = getExistingAccessor(i);
				if (sa==null) continue;
				AccessorReference childPath = AccessorReference.concatenate( path, new FieldIndexReference(i));
				sa.addListener(listener, cis, childPath);
			}
		}
	}
	
	@Override
	public void removeListener(Listener listener) throws AccessorException {
		ListenerEntry e = detachListener(listener);
		if (e==null) return;
		RecordInterestSet is = (RecordInterestSet) e.interestSet;
		
		// Apply component interest set to existing sub-accessors
		if (is.componentInterests!=null) {
			for (int i=0; i<count(); i++) {
				InterestSet cis = is.getComponentInterest(i);
				if (cis==null) continue;
				BinaryObject sa = getExistingAccessor(i);
				if (sa==null) continue;
				sa.removeListener(listener);
			}
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T extends Accessor> T getAccessor(AccessorReference reference) throws AccessorConstructionException {
		if (reference==null) return (T) this;
		
		if (reference instanceof LabelReference) {
			LabelReference lr = (LabelReference) reference;
			String fieldName = lr.label;
			Integer index = type().getComponentIndex(fieldName);
			if (index==null) throw new ReferenceException("RecordType doesn't have field by name \""+fieldName+"\"");
			BinaryObject sa = getFieldAccessor(index);
			if (reference.getChildReference() != null) sa = sa.getAccessor(reference.getChildReference());
			return (T) sa;			
		}		
		
		if (reference instanceof FieldIndexReference) {
			FieldIndexReference ref = (FieldIndexReference) reference;
			int index = ref.getIndex();
			BinaryObject sa = getFieldAccessor(index);
			if (reference.getChildReference() != null) sa = sa.getAccessor(reference.getChildReference());
			return (T) sa;			
		} 
		
		if (reference instanceof FieldNameReference) {
			FieldNameReference ref = (FieldNameReference) reference;
			String fieldName = ref.getFieldName();
			Integer index = type().getComponentIndex(fieldName);
			if (index==null) throw new ReferenceException("RecordType doesn't have field by name \""+fieldName+"\"");
			BinaryObject sa = getFieldAccessor(index);
			if (reference.getChildReference() != null) sa = sa.getAccessor(reference.getChildReference());
			return (T) sa;			
		} 
		
		throw new ReferenceException(reference.getClass()+" is not a subreference of RecordType");
		
	}

	@Override
	public void setValueNoflush(Binding binding, Object newValue) throws AccessorException {
		try {
			RecordBinding rb = (RecordBinding) binding;
			Serializer rs = rb.serializer();
			
			// Write
			long newSize = rs.getSize(newValue, null);
			b.setLength(newSize);
			b.position(0L);
			
			for (int index=0; index<count(); index++) {
				long pos = b.position();
				Binding cb = rb.getComponentBinding(index);
				Serializer cs = cb.serializer();
				Object cv = rb.getComponent(newValue, index);
				cs.serialize(b, cv);
				long len = b.position() - pos;
				BinaryObject sa = getExistingAccessor(index);
				if (sa!=null) {
					sa.b.setPositionInSource(pos, len);
				}
				
				// Notify field specific listeners
				if (listeners!=null) {
					ListenerEntry le = listeners;
					while (le!=null) {				
						RecordInterestSet is = le.getInterestSet();
						if (is.inNotificationsOf(index) /*&& !is.inNotifications()*/) {
							Event e = new ValueAssigned( new FieldIndexReference( index ), cb, is.inValuesOf(index) ? cv : null);
							le.emitEvent(e);
						}
						le = le.next;
					}				
				}
			}
			
			// Notify record specific listeners
			/*
			if (listeners!=null) {
				ListenerEntry le = listeners;
				while (le!=null) {				
					RecordInterestSet is = le.getInterestSet();
					if (is.inNotifications() && !is.inComponentNotifications()) {
						Event e = new ValueAssigned( binding, is.inValues() ? newValue : null);
						le.emitEvent(e);
					}
					le = le.next;
				}				
			}*/
			

		} catch (BindingException e) {
			throw new AccessorException(e);
		} catch (SerializationException e) {
			throw new AccessorException(e);
		} catch (IOException e) {
			throw new AccessorException(e);
		}
	}

	@Override
	Event applyLocal(Event e, boolean makeRollback) throws AccessorException {
		if (e instanceof ValueAssigned) {			
			ValueAssigned va = (ValueAssigned) e;			
			Event rollback = null; 
			if (makeRollback) {
				Binding binding = Bindings.getBinding(type());
				rollback = new ValueAssigned(binding, getValue(binding));				
			}
			setValueNoflush(va.newValue.getBinding(), va.newValue.getValue());
			return rollback;
		} else {
			throw new AccessorException("Cannot apply "+e.getClass().getName()+" to Record Type");
		}
	}
		

}

