/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.accessor.java;

import java.lang.ref.SoftReference;

import org.simantics.databoard.Bindings;
import org.simantics.databoard.accessor.Accessor;
import org.simantics.databoard.accessor.OptionalAccessor;
import org.simantics.databoard.accessor.error.AccessorConstructionException;
import org.simantics.databoard.accessor.error.AccessorException;
import org.simantics.databoard.accessor.error.ReferenceException;
import org.simantics.databoard.accessor.event.Event;
import org.simantics.databoard.accessor.event.OptionalValueAssigned;
import org.simantics.databoard.accessor.event.OptionalValueRemoved;
import org.simantics.databoard.accessor.event.ValueAssigned;
import org.simantics.databoard.accessor.impl.ListenerEntry;
import org.simantics.databoard.accessor.interestset.InterestSet;
import org.simantics.databoard.accessor.interestset.OptionalInterestSet;
import org.simantics.databoard.accessor.reference.AccessorReference;
import org.simantics.databoard.accessor.reference.LabelReference;
import org.simantics.databoard.accessor.reference.OptionalValueReference;
import org.simantics.databoard.adapter.AdaptException;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.OptionalBinding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.mutable.Variant;
import org.simantics.databoard.type.OptionalType;

public class JavaOptional extends JavaObject implements OptionalAccessor {

	/** Accessor to component */
	SoftReference<JavaObject> component;
	
	public JavaOptional(JavaObject parent, OptionalBinding binding, Object object) {
		super(parent, binding, object);
	}

	@Override
	public OptionalBinding getBinding() {
		return (OptionalBinding) binding;
	}

	@Override
	public OptionalType type() {
		return (OptionalType) getBinding().type();
	}	
	
	@SuppressWarnings("unchecked")
	@Override
	public <T extends Accessor> T getComponentAccessor() throws AccessorConstructionException
	{
		try {
			if (!getBinding().hasValue(object)) return null;
			
			// Get existing or create new
			JavaObject sa = getExistingAccessor();
			
			if (sa==null) {
				// Instantiate new accessor
				Binding cb = getBinding().getComponentBinding();
				Object cv = getBinding().getValue(object);
				
				// Instantiate correct sub accessor. 
				sa = createSubAccessor(this, cb, cv);
				component = new SoftReference<JavaObject>(sa);

				// Add listener to component, if it is in our interest set
				ListenerEntry le = listeners;
				while (le!=null) {				
					OptionalInterestSet is = le.getInterestSet();
					InterestSet cis = is.getComponentInterest(); 
					if (cis != null) {
						try {
							AccessorReference childPath = AccessorReference.concatenate( le.path, new OptionalValueReference() );
							sa.addListener(le.listener, cis, childPath);
						} catch (AccessorException e) {
							throw new AccessorConstructionException(e);
						}
					}
					le = le.next;
				}				
			}		
			
			return (T) sa;
		} catch (BindingException e) {
			throw new AccessorConstructionException(e);
		}
	}

	/**
	 * Get existing sub accessor
	 * 
	 * @return sub-accessor or <code>null</code>
	 */
	JavaObject getExistingAccessor()
	{
		// Get existing or create new
		SoftReference<JavaObject> ref = component;
		JavaObject accessor = (ref!=null)?(JavaObject)ref.get():null;
		return accessor;
	}	
	
	@Override
	public Object getComponentValue(Binding componentBinding)
			throws AccessorException {		
		try {
			if (!getBinding().hasValue(object))
				throw new AccessorException("There is no component value");
			
			Binding cb = getBinding().getComponentBinding();
			Object cv = getBinding().getValue(object);
			return Bindings.adapt(cv, cb, componentBinding);
		} catch (BindingException e) {
			throw new AccessorException(e);
		} catch (AdaptException e) {
			throw new AccessorException(e);
		}
	}
	
	@Override
	public void addListener(Listener listener, InterestSet interestSet,
			AccessorReference path) throws AccessorException {
		super.addListener(listener, interestSet, path);
		OptionalInterestSet is = (OptionalInterestSet) interestSet;
		InterestSet cis = is.getComponentInterest();
		if (cis==null) return;
		JavaObject sa = getExistingAccessor();
		if (sa==null) return;
		AccessorReference childPath = AccessorReference.concatenate( path, new OptionalValueReference() );
		sa.addListener(listener, cis, childPath);
	}
	
	@Override
	public void removeListener(Listener listener) throws AccessorException {
		ListenerEntry e = detachListener(listener);
		if (e==null) return;
		OptionalInterestSet is = (OptionalInterestSet) e.interestSet;
		
		InterestSet cis = is.getComponentInterest();
		if (cis==null) return;
		JavaObject sa = getExistingAccessor();
		if (sa==null) return;
		sa.removeListener(listener);		
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T extends Accessor> T getAccessor(AccessorReference reference) throws AccessorConstructionException {
		if (reference==null) return (T) this;

		if (reference instanceof LabelReference) {
			LabelReference lr = (LabelReference) reference;
			if (lr.label.equals("o")) {
				Accessor result = getComponentAccessor();
				if (reference.getChildReference() != null)
					result = result.getAccessor(reference.getChildReference());
				return (T) result;			
			}			
		}
		
		if (reference instanceof OptionalValueReference) {
			Accessor result = getComponentAccessor();
			if (reference.getChildReference() != null)
				result = result.getAccessor(reference.getChildReference());
			return (T) result;			
		} 
		
		throw new ReferenceException(reference.getClass()+" is not a reference of OptionalType");		
	}

	@Override
	public void setValue(Binding binding, Object newValue) throws AccessorException {
		try {
			OptionalBinding rb = (OptionalBinding) binding;
			Object rv = newValue;
			boolean rvHasValue = rb.hasValue(newValue);
			
			if (rvHasValue) { 
				Binding rcb = rb.getComponentBinding();
				Object rcv = rb.getValue(rv);
				setComponentValue(rcb, rcv);
			} else {
				setNoValue();				
			}
			
		} catch (BindingException e) {
			throw new AccessorException(e);
		}
	}

	@Override
	public boolean hasValue() throws AccessorException {
		try {
			return getBinding().hasValue(object);
		} catch (BindingException e) {
			throw new AccessorException(e);
		}
	}


	@Override
	public void setNoValue() throws AccessorException {
		try {
			boolean hadValue = getBinding().hasValue(object);
			if (!hadValue) return;
			
			// Write
			getBinding().setNoValue(object);			

			// Disconnect sub-accessor
			JavaObject sa = getExistingAccessor();
			component = null;
			// Notify about disconnection of sub-accessor
			if (sa!=null) sa.invalidatedNotification();
						
			// Notify Listeners
			ListenerEntry le = listeners;
			while (le!=null) {				
				OptionalInterestSet is = le.getInterestSet();
				if (is.inNotifications()) {
					OptionalValueRemoved e = new OptionalValueRemoved();
					le.emitEvent(e);
				}
				le = le.next;
			}
			
		} catch (BindingException e) {
			throw new AccessorException(e);
		}
	}

	@Override
	public void setComponentValue(Binding binding, Object value)
			throws AccessorException {
		try {
			Binding rcb = binding;
			Object rcv = value;
			boolean hadValue = getBinding().hasValue(object);

			// Write to component
			JavaObject sa = getExistingAccessor();
			if (sa!=null) {
				assert(hadValue);
				sa.setValue(rcb, rcv);
				return;
			}

			Binding lcb = getBinding().getComponentBinding();
			Object lcv = Bindings.adapt(rcv, rcb, lcb);			
			
			// Write
			getBinding().setValue(object, lcv);
			
			// Notify Listeners
			ListenerEntry le = listeners;
			while (le!=null) {				
				OptionalInterestSet is = le.getInterestSet();
				if (is.inNotifications()) {					
					Variant newComponentValue = null;
					if (is.inValues()) newComponentValue = new Variant(lcb, lcb.isImmutable() ? lcv : lcb.clone(lcv));
					OptionalValueAssigned e = new OptionalValueAssigned(newComponentValue);
					le.emitEvent(e);
				}
				
				// Attach component interest
//				if (is.getComponentInterest()!=null && getComponentAccessor()!=null) {
//					sa = getComponentAccessor();
//				}
				
				le = le.next;
			}
			
		} catch (BindingException e) {
			throw new AccessorException(e);
		} catch (AdaptException e) {
			throw new AccessorException(e);
//		} catch (AccessorConstructionException e) {
//			throw new AccessorException(e);
		}
		
	}

	@Override
	Event applyLocal(Event e, boolean makeRollback) throws AccessorException {
		try {
			Event rollback = null;
			if (makeRollback) {
				OptionalBinding b = getBinding();
				Binding cb = getBinding().getComponentBinding();			
				rollback = hasValue() ? new OptionalValueAssigned(cb, b.getValue(object)) : new OptionalValueRemoved(); 
			}			
			if (e instanceof ValueAssigned) {
				ValueAssigned va = (ValueAssigned) e;
				setValue(va.newValue.getBinding(), va.newValue.getValue());
				return rollback;
			} else 		
			if (e instanceof OptionalValueAssigned) {
				OptionalValueAssigned oa = (OptionalValueAssigned) e;
				if (oa.newValue==null) throw new AccessorException("Cannot apply field assignment event, the value is missing");
				setComponentValue(oa.newValue.getBinding(), oa.newValue.getValue());
			} else if (e instanceof OptionalValueRemoved) {
				//OptionalValueRemoved or = (OptionalValueRemoved) e;
				setNoValue();
			} else throw new AccessorException("Unexpected event type "+e.getClass().getName()+" for an Optional Type");
			
			return rollback;
		} catch (BindingException be) {
			throw new AccessorException( be );
		}
	}

	
}

