/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.accessor.reference;

import java.io.IOException;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.simantics.databoard.Bindings;
import org.simantics.databoard.adapter.AdaptException;
import org.simantics.databoard.annotations.Optional;
import org.simantics.databoard.annotations.Union;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.mutable.Variant;
import org.simantics.databoard.binding.mutable.VariantContainerBinding;
import org.simantics.databoard.history.StringVariantBinding;
import org.simantics.databoard.serialization.SerializationException;
import org.simantics.databoard.util.URIUtil;

/**
 * AccessorRef is a reference to a child node in the tree representation of 
 * a container. <p>
 * 
 * ComponentReference[] is a path from a node to a child or decendent node 
 * in the tree representation of the container. <p>   
 *  
 * Reference has three serializable formats: Binary, URL compatible Path.
 * 
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
@Union({
		ArrayIndexReference.class,
		KeyReference.class, 
		OptionalValueReference.class,
		FieldIndexReference.class,
		FieldNameReference.class,
		TagReference.class,
		TagIndexReference.class,
		TagNameReference.class,
		VariantValueReference.class, 
		LabelReference.class
	})
public abstract class AccessorReference implements Cloneable {

	/**
	 * Get reference path from AccessorReference path.
	 * <a href="https://www.simantics.org/wiki/index.php/Databoard_Specification#Accessor_Reference">Accessor Reference</a>
	 * 
	 * @param path
	 * @return reference path or <code>null</code> if there is no path
	 */
	public static AccessorReference parseString(String path) {
		StringTokenizer st = new StringTokenizer(path, "/", false);
		if (!st.hasMoreTokens()) return null;
		AccessorReference first = createSingleReference( st.nextToken() );		
		AccessorReference ref = first;
		while (st.hasMoreTokens()) {
			ref.childReference = createSingleReference(st.nextToken());
			ref = ref.childReference;
		}
		
		return first;
	}
	
	public static AccessorReference parseBinary(byte[] binaryRef) 
	throws SerializationException 
	{
		try {
			Binding binding = Bindings.getBindingUnchecked(AccessorReference.class);
			AccessorReference result;
			result = (AccessorReference) binding.serializer().deserialize(binaryRef);
			return result;
		} catch (IOException e) {
			throw new SerializationException(e);
		} catch (BindingException e) {
			throw new SerializationException(e);
		}
	}
	
	public final static Pattern ARRAY_ELEMENT_PATTERN = Pattern.compile("i-(\\d*)");
	public final static Pattern MAP_PATTERN = Pattern.compile("m-(\\p{ASCII}*)");
	public final static Pattern FIELD_INDEX_PATTERN = Pattern.compile("f-(\\d)*");
	public final static Pattern FIELD_NAME_PATTERN = Pattern.compile("n-(\\p{ASCII}*)");
	public final static Pattern UNION_INDEX_PATTERN = Pattern.compile("u-(\\d*)");
	public final static Pattern UNION_NAME_PATTERN = Pattern.compile("t-(\\p{ASCII}*)*");
	
	public @Optional AccessorReference childReference;
	
	protected AccessorReference() {}
	
	protected AccessorReference(AccessorReference childReference) {
		this.childReference = childReference;
	}
	
	public AccessorReference getChildReference() {
		return childReference;
	}
	
	public boolean hasChildReference() {
		return childReference != null;
	}
		
	public void setChildReference(AccessorReference childReference) {
		this.childReference = childReference;
	}

	public String toPath() {
		return toPath(true);
	}
	
	/**
	 * Converts the reference path into string representation.
	 * 
	 * @param labelReference if true return label references.
	 * @return path string representation
	 */
	public String toPath(boolean labelReference) {
		if (childReference == null) return toString();
		StringBuilder sb = new StringBuilder();
		AccessorReference ref = this;
		while (ref!=null) {
			if (sb.length() > 0) sb.append("/");
			sb.append( ref.toString(labelReference) );
			ref = ref.getChildReference();
		}
		return sb.toString();
	}
	
	/**
	 * Convert the reference into its string representation
	 * 
	 * @return reference string representation
	 */
	public String toString() {
		return toString(true);
	}
	
	/**
	 * Convert the reference into string representation.<p>
	 *
	 * If <code>labelReference</code> is true, the string representation is
	 * more user readable but has weaker typing. It serializes into
	 * instances of LabelReference.
	 * 
	 * For instance Record Field Reference is "n-Children", but label reference "Children".
	 * 
	 * Some references cannot be converted into LabelReference. 
	 * E.g. string representation of FieldNameReference("i-5") is ambiguous with ArrayIndexReference(5). 
	 * 
	 * @param labelReference if true returns  
	 * @return string representation 
	 */
	public abstract String toString(boolean labelReference); 

	public abstract AccessorReference clone();
	
	public AccessorReference tail() {
		AccessorReference result = this;
		while (result.childReference!=null) result = result.childReference;
		return result;
	}
	
	/**
	 * Create a concatenation of two references. Prefix part is cloned,
	 * suffix is linked.
	 * 
	 * @param pathToBeCloned prefix path, or <tt>null</tt>
	 * @param ref suffix path, or <tt>null</tt>
	 * @return path
	 */
	public static AccessorReference concatenate(AccessorReference pathToBeCloned, AccessorReference ref) {
		if (pathToBeCloned==null) return ref;
		AccessorReference result = pathToBeCloned.clone();
		if (ref==null) return result;
		result.tail().setChildReference(ref);
		return result;
	}
	
	/**
	 * Creates a compilation of individual a references into a one refence. 
	 *  
	 * @param refs
	 * @return reference or <code>null</code> if there are no elements
	 */
	public static AccessorReference compile(AccessorReference ... refs) {
		if (refs.length==0) return null;
		AccessorReference first = refs[0].clone();
		AccessorReference r = first;
		for (int i=1; i<refs.length; i++) {
			AccessorReference next = refs[i].clone();
			r.setChildReference( next );
			r = next;
			while(r.childReference!=null) r = r.childReference;
		}
		return first;
	}
	
	/**
	 * Create accessor reference from string representation.
	 * This doesn't parse path separators.
	 * 
	 * @see https://www.simantics.org/wiki/index.php/Databoard_Specification#Accessor_Reference
	 * 
	 * @param ref
	 * @return
	 */
	static AccessorReference createSingleReference(String ref) {
		Matcher m;
		
		m = ARRAY_ELEMENT_PATTERN.matcher( ref );
		if (m.matches()) {
			return new FieldIndexReference( Integer.parseInt( m.group(1) ) );
		}
		
		m = MAP_PATTERN.matcher( ref );
		if (m.matches()) {
			String keyStr = m.group(1);
			Variant key;
			try {
				key = (Variant) Bindings.adapt(keyStr, StringVariantBinding.INSTANCE, VariantContainerBinding.INSTANCE);
			} catch (AdaptException e) {
				throw new IllegalArgumentException("Not string variant "+ref, e);
			}
			return new KeyReference(key);
		}
		
		m = FIELD_INDEX_PATTERN.matcher( ref );
		if (m.matches()) {
			return new FieldIndexReference( Integer.parseInt( m.group(1) ) );
		}
		
		m = FIELD_NAME_PATTERN.matcher( ref );
		if (m.matches()) {
			String encoded = m.group(1);
			String name = URIUtil.decodeURI(encoded);
			return new FieldNameReference( name );
		}
		
		m = UNION_INDEX_PATTERN.matcher( ref );
		if (m.matches()) {
			return new TagIndexReference( Integer.parseInt( m.group(1) ) );
		}
		
		m = UNION_NAME_PATTERN.matcher( ref );
		if (m.matches()) {
			String encoded = m.group(1);
			String name = URIUtil.decodeURI(encoded);
			return new TagNameReference( name );
		}

		if (ref.equals("uv")) {
			return new TagReference();
		}
		
		if (ref.equals("o")) {
			return new OptionalValueReference();
		}
		
		if (ref.equals("v")) {
			return new VariantValueReference();
		}
		
		String text = URIUtil.decodeURI( ref );
		return new LabelReference( text );		
	}
	
}

