/*******************************************************************************
 * Copyright (c) 2010- Association for Decentralized Information Management in
 * Industry THTH ry.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *    VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.datasource;

import java.util.Collection;
import java.util.Map;

import org.simantics.databoard.binding.mutable.Variant;
import org.simantics.databoard.type.DataType;

/**
 * Interface for stepwise data sources.
 * 
 * The values of a Stepwise Datasource changes only at #step().  
 * 
 * Root node is mandatory. The identifier of root node is the an empty record.
 * You can instantiate root node id with ID_BINDING.createDefaultUnchecked()
 * 
 * @see ExampleDatasource 
 * @see StepwiseDatasourceAccessor
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public interface IStepwiseDatasource {
	
	/**
	 * Get a list of all node identifiers. 
	 * 
	 * @param list to be filled with identifiers.
	 */
	void getAll(Collection<Variant> list);
	
	/**
	 * Verifies that the datasource has a node
	 * 
	 * @param nodeId
	 * @return <code>true</code> if node exists, false if not
	 */
	boolean hasNode(Variant nodeId);
	
	/**
	 * Get children of a node 
	 * 
	 * @param nodeId
	 * @param children a collection to be filled with children
	 * @throws DatasourceException if the node does not exist
	 */
	void getChildren(Variant nodeId, Collection<Variant> children) throws DatasourceException;
	
	/**
	 * Get value of a node
	 * 
	 * @param nodeId
	 * @param value variant where the value is written to
	 * @return false if there is no value, otherwise true
	 * @throws DatasourceException if node does not exist
	 */
	boolean getValue(Variant nodeId, Variant value) throws DatasourceException;
	
	/**
	 * Get the value type of a node.
	 * 
	 * @param nodeId
	 * @return value type or <code>null</code> if there is no value
	 * @throws DatasourceException if the node does not exist
	 */
	DataType getNodeType(Variant nodeId) throws DatasourceException;
		
	/**
	 * Go to the next step of value.
	 * 
	 * In case of exception, the state of the values is undetermined. The values
	 * may be from previous step, or from the failed step.
	 * 
	 * @param monitorSet a collection of nodes that are monitored for change, or null
	 * @return a list of nodes whose value changed.
	 * @throws DatasourceException if error occurs
	 */
	Collection<Variant> step(Collection<Variant> monitorSet) throws DatasourceException;
	
	/**
	 * Get all labels of a node. 
	 * Locale is ISO-639 coded. Label for locale "en" is always available.
	 * 
	 * See https://www.simantics.org/wiki/index.php/Databoard_Specification#Utility_types for details about localized text.   
	 * 
	 * @param nodeId
	 * @param locale ISO-639 coded locale.
	 * @return the label or <code>null</code> if there is no label for the locale. "en" is available.
	 * @throws DatasourceException if error occurs
	 */
	String getLabel(Variant nodeId, String locale) throws DatasourceException;
	
	/**
	 * Get all labels of a node. <tt>locales</tt> map is filled with locale specific labels.
	 * Locale is ISO-639 encoded language/region code.
	 * Label for "en" is always available.
	 * 
	 * @param nodeId
	 * @param labels a collection to be filled
	 * @throws DatasourceException if error occurs
	 */
	void getLabels(Variant nodeId, Map<String, String> labels) throws DatasourceException;
	
	/**
	 * Add datamodel listener
	 * 
	 * @param listener
	 */
	void addListener(DatasourceListener listener);
	
	/**
	 * Remove datamodel listener
	 * 
	 * @param listener
	 */
	void removeListener(DatasourceListener listener);
	
	public interface DatasourceListener {
		void onStep();
//		void onNodeAdded(Variant nodeId);
//		void onNodeRemoved(Variant nodeId);
//		void onLabelChanged(Variant nodeId);
	}
	
}

