/*******************************************************************************
 * Copyright (c) 2010- Association for Decentralized Information Management in
 * Industry THTH ry.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *    VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.datasource;

import java.lang.ref.SoftReference;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import org.simantics.databoard.Accessors;
import org.simantics.databoard.Bindings;
import org.simantics.databoard.accessor.Accessor;
import org.simantics.databoard.accessor.MapAccessor;
import org.simantics.databoard.accessor.RecordAccessor;
import org.simantics.databoard.accessor.error.AccessorConstructionException;
import org.simantics.databoard.accessor.error.AccessorException;
import org.simantics.databoard.accessor.error.ReferenceException;
import org.simantics.databoard.accessor.error.WriteAccessException;
import org.simantics.databoard.accessor.event.Event;
import org.simantics.databoard.accessor.event.MapEntryAdded;
import org.simantics.databoard.accessor.event.MapEntryRemoved;
import org.simantics.databoard.accessor.event.ValueAssigned;
import org.simantics.databoard.accessor.impl.CompositeRecord;
import org.simantics.databoard.accessor.impl.ListenerEntry;
import org.simantics.databoard.accessor.interestset.InterestSet;
import org.simantics.databoard.accessor.interestset.MapInterestSet;
import org.simantics.databoard.accessor.java.JavaOptional;
import org.simantics.databoard.accessor.reference.AccessorReference;
import org.simantics.databoard.accessor.reference.KeyReference;
import org.simantics.databoard.accessor.reference.LabelReference;
import org.simantics.databoard.adapter.AdaptException;
import org.simantics.databoard.adapter.Adapter;
import org.simantics.databoard.adapter.AdapterConstructionException;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.MapBinding;
import org.simantics.databoard.binding.RecordBinding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.java.ArrayListBinding;
import org.simantics.databoard.binding.java.StringBindingDefault;
import org.simantics.databoard.binding.java.TreeMapBinding;
import org.simantics.databoard.binding.mutable.ContainerOptionalBinding;
import org.simantics.databoard.binding.mutable.ValueContainer;
import org.simantics.databoard.binding.mutable.Variant;
import org.simantics.databoard.binding.mutable.VariantContainerBinding;
import org.simantics.databoard.datasource.IStepwiseDatasource.DatasourceListener;
import org.simantics.databoard.history.StringVariantBinding;
import org.simantics.databoard.type.MapType;

/**
 * Adapt Datasource into a read-only accessor.
 * 
 * This implementation does not support modifications to the datamodel and 
 * labels of the data source. 
 * 
 * This is very simple but not the most optimal implementation. 
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class StepwiseDatasourceAccessor extends CompositeRecord {

	// Bindings for node fields
	static Binding VARIANT_BINDING = VariantContainerBinding.INSTANCE;
	static Binding LOCALIZEDTEXT_BINDING = new TreeMapBinding( StringBindingDefault.INSTANCE, StringBindingDefault.INSTANCE );
	static Binding VARIANT_ARRAY_BINDING = new ArrayListBinding( VariantContainerBinding.INSTANCE );
	static Binding OPTIONAL_VARIANT_BINDING = new ContainerOptionalBinding( VariantContainerBinding.INSTANCE );
	static Binding NODE_BINDING = Bindings.getBindingUnchecked( NodeRecord.class );
	static MapType NODES_TYPE;
	
	IStepwiseDatasource datasource;
	NodesMapAccessor nodesAccessor;
	
	public StepwiseDatasourceAccessor(IStepwiseDatasource datasource) {
		this.datasource = datasource;
		
		addField("nodes", nodesAccessor = new NodesMapAccessor());		
		datasource.addListener( nodesAccessor.listener );
	}
	
	public void dispose() {
		datasource.removeListener( nodesAccessor.listener );
	}
	
	static {
		NODES_TYPE = new MapType( VARIANT_BINDING.type(), NODE_BINDING.type() );
	}
	
	class NodeAccessor extends CompositeRecord {
		
		Variant nodeId;
		Variant wrappedNodeId;
		Accessor idAccessor;
		Accessor labelsAccessor;
		Accessor childrenAccessor;
		JavaOptional valueAccessor;
		
		public NodeAccessor(Variant nodeId) throws AccessorConstructionException, DatasourceException {
			this.nodeId = nodeId;
			this.wrappedNodeId = new Variant( VARIANT_BINDING, nodeId );
			
			// Identifier
			idAccessor = Accessors.getAccessor( VARIANT_BINDING, nodeId );
			
			// Read labels
			TreeMap<String, String> labels = new TreeMap<String, String>();
			datasource.getLabels(nodeId, labels);
			labelsAccessor = Accessors.getAccessor( LOCALIZEDTEXT_BINDING, labels );
			
			// Read children
			ArrayList<Variant> children = new ArrayList<Variant>();
			datasource.getChildren(nodeId, children);
			childrenAccessor = Accessors.getAccessor(VARIANT_ARRAY_BINDING, children);
			
			// Value Accessor
			ValueContainer container = new ValueContainer();
			Variant value = new Variant();
			// Read value
			if (datasource.getValue(nodeId, value)) container.setValue( value );
			valueAccessor = Accessors.getAccessor(OPTIONAL_VARIANT_BINDING, container);
			
			addField("id", idAccessor);
			addField("labels", labelsAccessor);
			addField("children", childrenAccessor);
			addField("value", valueAccessor);			
		}
		
		void setValue(Variant newValue) {
			try {
				valueAccessor.setValue(VARIANT_BINDING, newValue);
			} catch (AccessorException e) {
				throw new RuntimeException( e );
			}
		}
	}
	
	/**
	 * Accessor to datasource nodes. Nodes are read from a IStepwiseDatasource.
	 * 
	 * Sub-accessor for a node is instantiated on demand and stored with a soft
	 * reference.
	 *
	 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
	 */
	class NodesMapAccessor implements MapAccessor {
		/** Accessors to children */
		TreeMap<Variant, SoftReference<NodeAccessor>> children = new TreeMap<Variant, SoftReference<NodeAccessor>>( VARIANT_BINDING );

		TreeSet<Variant> ids;
		
		/** Listeners */
		ListenerEntry listeners = null;		
		
		DatasourceListener listener = new DatasourceListener() {
			@Override
			public void onStep() {
				for (SoftReference<NodeAccessor> ref : children.values()) {
					NodeAccessor na = ref.get();
					if (na == null) continue;
					
					try {
						// which nodes changed
						Variant newValue = new Variant(); 
						boolean hasValue = datasource.getValue(na.nodeId, newValue);
						if (!hasValue) {
							na.valueAccessor.setNoValue();
						} else {
							na.valueAccessor.setComponentValue(VARIANT_BINDING, newValue);
						}
					} catch (DatasourceException e) {
						e.printStackTrace();
					} catch (AccessorException e) {
						e.printStackTrace();
					}
				}
			}
		};
		
		@Override
		public MapType type() {
			return NODES_TYPE;
		}
		
		TreeSet<Variant> getIds() {
			if (ids==null) { 
				ids = new TreeSet<Variant>( VARIANT_BINDING );
				datasource.getAll(ids);
			}
			return ids;
		}
		
		/**
		 * Get existing sub accessor
		 * @param index
		 * @return sub-accessor or <code>null</code>
		 */
		NodeAccessor getExistingAccessor(Variant nodeId)
		{
			SoftReference<NodeAccessor> ref = children.get(nodeId);
			if (ref==null) return null;
			NodeAccessor result = ref.get();
			if (result==null) {
				children.remove(nodeId);
				return null;
			}
			return result;
		}	
		
		@Override
		public <T extends Accessor> T getAccessor(AccessorReference reference)
				throws AccessorConstructionException {
			if (reference==null) return (T) this;
			if (reference instanceof LabelReference) {
				LabelReference lr = (LabelReference) reference;
				try {
					Variant value = (Variant) Bindings.adapt(lr.label, StringVariantBinding.INSTANCE, VARIANT_BINDING);
					Accessor result = (T) getValueAccessor(VARIANT_BINDING, value);
					if (reference.getChildReference() != null)
						result = result.getAccessor(reference.getChildReference());
					return (T) result;			
				} catch (AdaptException e) {
					throw new ReferenceException(e);				
				}			
			} else if (reference instanceof KeyReference) {
				KeyReference ref = (KeyReference) reference;
				Accessor result = getValueAccessor(VARIANT_BINDING, ref.key);
				if (reference.getChildReference() != null)
					result = result.getAccessor(reference.getChildReference());
				return (T) result;			
			} 
			throw new ReferenceException(reference.getClass().getName()+" is not a reference of a map");	
		}

		@Override
		public <T extends Accessor> T getValueAccessor(Binding keyBinding,
				Object key) throws AccessorConstructionException {
			try {
				Variant nodeId = (Variant) Bindings.adapt(key, keyBinding, VARIANT_BINDING);
				if (!datasource.hasNode(nodeId)) throw new AccessorConstructionException("Map doesn't contain the requested element");				
				
				NodeAccessor sa = getExistingAccessor(nodeId);
				if (sa!=null) return (T) sa;
				
				// Create NodeAccessor
				sa = new NodeAccessor(nodeId);
				children.put(nodeId, new SoftReference<NodeAccessor>(sa));
				
				// Add component interest sets
				ListenerEntry le = listeners;
				if (le!=null) {
					
					while (le!=null) {				
						MapInterestSet is = le.getInterestSet();
	
						// Generic element interest
						InterestSet gis = is.getComponentInterest(); 
						if (gis != null) {
							try {							
								AccessorReference childPath = AccessorReference.concatenate(le.path, new KeyReference( sa.wrappedNodeId ) );
								sa.addListener(le.listener, gis, childPath);
							} catch (AccessorException e) {
								throw new AccessorConstructionException(e);
							}
						}
							
						// Specific element interest
						InterestSet cis = is.getComponentInterest( sa.wrappedNodeId ); 
						if (cis != null) {
							try {
								AccessorReference childPath = AccessorReference.concatenate(le.path, new KeyReference( sa.wrappedNodeId ) );
								sa.addListener(le.listener, cis, childPath);
							} catch (AccessorException e) {
								throw new AccessorConstructionException(e);
							}
						}
						
						// Next listener
						le = le.next;
					}
				}
				
				return (T) sa;
			} catch (AdaptException e) {
				throw new AccessorConstructionException(e);
			} catch (DatasourceException e) {
				throw new AccessorConstructionException(e);
			}	
		}
		
		@Override
		public void addListener(Listener listener, InterestSet interestSet, AccessorReference path) throws AccessorException {
			listeners = ListenerEntry.link(listeners, listener, interestSet, path);
			MapInterestSet is = (MapInterestSet) interestSet;
			
			// Apply listener to existing sub-accessors
			for (Variant nodeId : children.keySet()) {
				NodeAccessor sa = getExistingAccessor(nodeId);
				if (sa==null) continue;
				
				InterestSet cis = is.getComponentInterest();
				if (cis!=null) {
					AccessorReference childPath = AccessorReference.concatenate( path, new KeyReference(sa.wrappedNodeId) );
					sa.addListener(listener, cis, childPath);				
				}
				cis = is.getComponentInterest( sa.wrappedNodeId );
				if (cis!=null) {
					AccessorReference childPath = AccessorReference.concatenate( path, new KeyReference(sa.wrappedNodeId) );
					sa.addListener(listener, cis, childPath);				
				}
			}			
		}

		protected ListenerEntry detachListener(Listener listener) throws AccessorException {
			ListenerEntry e = listeners;
			ListenerEntry p = null;
			while (e!=null) {
				// Found match
				if (e.listener == listener) {
					// The match was the first entry of the linked list
					if (p==null) {
						listeners = e.next;
						return e;
					}
					// Some other entry, unlink e
					p.next = e.next;
					return e;
				}
				p = e;
				e = e.next;
			}
			return null;		
		}
		
		@Override
		public void removeListener(Listener listener) throws AccessorException {
			ListenerEntry e = detachListener(listener);
			if (e==null) return;
			MapInterestSet is = (MapInterestSet) e.interestSet;
						
			for (Entry<Variant, SoftReference<NodeAccessor>> entry : children.entrySet()) {
				NodeAccessor sa = entry.getValue().get();
				if (sa==null) continue;
				
				InterestSet cis = is.getComponentInterest();
				if (cis!=null) {
					sa.removeListener(listener);				
				}
				cis = is.getComponentInterest( sa.wrappedNodeId );
				if (cis!=null) {
					sa.removeListener(listener);				
				}
			}
		}

		@Override
		public int size() throws AccessorException {
			TreeSet<Variant> ids = getIds();
			return ids.size();
		}

		@Override
		public Object getValue(Binding binding) throws AccessorException {
			if (!binding.type().equals(NODES_TYPE)) throw new AccessorException("Wrong binding");
			
			try {
				MapBinding mb = (MapBinding) binding;
				
				TreeSet<Variant> ids = getIds();
				int size = ids.size();
				
				Variant keys[] = ids.toArray( new Variant[size] );
				Object values[] = new NodeRecord[ size ];
				
				for (int i=0; i<size; i++) {
					Variant key = keys[i];
					NodeRecord record = new NodeRecord();
					record.id = key;
					record.labels = new TreeMap<String, String>();
					record.value = new Variant();
					record.children = new ArrayList<Variant>(1);
					
					datasource.getLabels(key, record.labels);
					datasource.getChildren(key, record.children);
					datasource.getValue(key, record.value);
					
					values[i] = record;
				}
				
				return mb.create(keys, values);
			} catch (DatasourceException e) {
				throw new AccessorException(e);
			} catch (BindingException e) {
				throw new AccessorException(e);
			}
		}
		
		@Override
		public Object get(Binding keyBinding, Object key, Binding valueBinding)
				throws AccessorException {
			try {
				Variant nodeId = (Variant) Bindings.adapt(key, keyBinding, VARIANT_BINDING);
				if (!datasource.hasNode(nodeId)) return null;
				RecordBinding rb = (RecordBinding) valueBinding;
				
				TreeMap<String, String> labels = new TreeMap<String, String>();
				Variant value = new Variant();
				ValueContainer container = new ValueContainer();
				ArrayList<Variant> children = new ArrayList<Variant>();

				
				datasource.getLabels(nodeId, labels);
				datasource.getChildren(nodeId, children);
				if (datasource.getValue(nodeId, value)) container.value = value;
				
				Object nodeId2 = Bindings.adapt(nodeId, VARIANT_BINDING, rb.getComponentBinding(0));
				Object labels2 = Bindings.adapt(labels, LOCALIZEDTEXT_BINDING, rb.getComponentBinding(1)); 
				Object children2 = Bindings.adapt(labels, VARIANT_ARRAY_BINDING, rb.getComponentBinding(2));
				Object value2  = Bindings.adapt(container, OPTIONAL_VARIANT_BINDING, rb.getComponentBinding(3));
				
				return rb.create(nodeId2, labels2, children2, value2);
			} catch (AdaptException e) {
				throw new AccessorException(e);
			} catch (DatasourceException e) {
				throw new AccessorException(e);
			} catch (BindingException e) {
				throw new AccessorException(e);
			}
			
		}

		@Override
		public boolean containsKey(Binding keyBinding, Object key)
				throws AccessorException {
			try {
				Variant nodeId = (Variant) Bindings.adapt(key, keyBinding, VARIANT_BINDING);
				return datasource.hasNode(nodeId);
			} catch (AdaptException e) {
				throw new AccessorException( e );
			}
		}

		@Override
		public boolean containsValue(Binding valueBinding, Object value)
				throws AccessorException {
			if (!valueBinding.type().equals( NODE_BINDING.type() )) {
				throw new AccessorException("Unexpected value type");
			}
			
			try {
				RecordAccessor ra = (RecordAccessor) Accessors.getAccessor(valueBinding, value);
				Variant nodeId = (Variant) ra.getFieldValue(0, VARIANT_BINDING);
				if (!datasource.hasNode(nodeId)) return false;
				
				// Verify fields match
				Object value2 = ra.getValue(valueBinding);
				return valueBinding.equals(value, value2);
			} catch (AccessorConstructionException e) {
				throw new AccessorException( e );
			}
		}

		@Override
		public void setValue(Binding mapBinding, Object newMap)
				throws AccessorException {
			throw new WriteAccessException();
		}

		@Override
		public void put(Binding keyBinding, Object key, Binding valueBinding,
				Object value) throws AccessorException {
			throw new WriteAccessException();
		}

		@Override
		public void remove(Binding keyBinding, Object key)
				throws AccessorException {
			throw new WriteAccessException();
		}

		@Override
		public void clear() throws AccessorException {
			throw new WriteAccessException();
		}

		@Override
		public void putAll(Binding keyBinding, Binding valueBinding,
				Map<Object, Object> from) throws AccessorException {
			throw new WriteAccessException();
		}

		@Override
		public void putAll(Binding keyBinding, Binding valueBinding,
				Object[] keys, Object[] values) throws AccessorException {
			throw new WriteAccessException();
		}

		@Override
		public void getAll(Binding keyBinding, Binding valueBinding,
				Map<Object, Object> to) throws AccessorException {
			Set<Variant> ids = getIds();
			int size = ids.size();
			
			try {
				Adapter keyAdapter = Bindings.getAdapter(VARIANT_BINDING, keyBinding);
				Iterator<Variant> iter = ids.iterator(); 
				for (int i=0; i<size; i++) {
					Variant id = iter.next();
					Object value = get(VARIANT_BINDING, id, valueBinding);
					Object key = keyAdapter.adapt( iter.next() );
					to.put(key, value);
				}	
			} catch (AdapterConstructionException e) {
				throw new AccessorException(e);
			} catch (AdaptException e) {
				throw new AccessorException(e);
			}
		}

		@Override
		public void getAll(Binding keyBinding, Binding valueBinding, Object[] keys, Object[] values) throws AccessorException {
			Set<Variant> ids = getIds();
			int size = ids.size();
			if (keys.length<size || values.length<size) throw new AccessorException("array too small");
			
			try {
				Adapter keyAdapter = Bindings.getAdapter(VARIANT_BINDING, keyBinding);
				Iterator<Variant> iter = ids.iterator(); 
				for (int i=0; i<size; i++) {
					Variant id = iter.next();
					values[i] = get(VARIANT_BINDING, id, valueBinding);
					keys[i] = keyAdapter.adapt( iter.next() );
				}	
			} catch (AdapterConstructionException e) {
				throw new AccessorException(e);
			} catch (AdaptException e) {
				throw new AccessorException(e);
			}
		}

		@Override
		public Object[] getKeys(Binding keyBinding) throws AccessorException {
			try {
				Set<Variant> ids = getIds();
				int size = ids.size();
				Object[] result = new Object[ size ];
				Adapter keyAdapter = Bindings.getAdapter(VARIANT_BINDING, keyBinding);
				Iterator<Variant> iter = ids.iterator(); 
				for (int i=0; i<size; i++) 
					result[i] = keyAdapter.adapt( iter.next() );
				return result;
			} catch (AdapterConstructionException e) {
				throw new AccessorException(e);
			} catch (AdaptException e) {
				throw new AccessorException(e);
			}
		}

		@Override
		public Object[] getValues(Binding valueBinding)
				throws AccessorException {
			Set<Variant> ids = getIds();
			int size = ids.size();
			Object[] result = new Object[ size ];
			Iterator<Variant> iter = ids.iterator(); 
			for (int i=0; i<size; i++) {
				Variant id = iter.next();
				result[i] = get(VARIANT_BINDING, id, valueBinding);
			}			
			
			return result;
		}

		@Override
		public Object getFirstKey(Binding keyBinding) throws AccessorException {
			try {
				Variant firstId = getIds().first();
				if (firstId==null) return null;			
				return Bindings.adapt(firstId, VARIANT_BINDING, keyBinding);
			} catch (AdaptException e) {
				throw new AccessorException( e );
			}
		}

		@Override
		public Object getLastKey(Binding keyBinding) throws AccessorException {
			try {
				Variant lastId = getIds().last();
				if (lastId==null) return null;			
				return Bindings.adapt(lastId, VARIANT_BINDING, keyBinding);
			} catch (AdaptException e) {
				throw new AccessorException( e );
			}
		}

		@Override
		public Object getLowerKey(Binding keyBinding, Object key)
				throws AccessorException {
			try {
				Variant nodeId = (Variant) Bindings.adapt(key, keyBinding, VARIANT_BINDING);
				TreeSet<Variant> ids = getIds();
				Variant lowerId = ids.lower(nodeId);
				if (lowerId == null) return null;				
				return Bindings.adapt(lowerId, VARIANT_BINDING, keyBinding);
			} catch (AdaptException e) {
				throw new AccessorException( e );
			}
		}

		@Override
		public Object getFloorKey(Binding keyBinding, Object key)
				throws AccessorException {
			try {
				Variant nodeId = (Variant) Bindings.adapt(key, keyBinding, VARIANT_BINDING);
				TreeSet<Variant> ids = getIds();
				Variant floorId = ids.floor(nodeId);
				if (floorId == null) return null;				
				return Bindings.adapt(floorId, VARIANT_BINDING, keyBinding);
			} catch (AdaptException e) {
				throw new AccessorException( e );
			}
		}

		@Override
		public Object getCeilingKey(Binding keyBinding, Object key)
				throws AccessorException {
			try {
				Variant nodeId = (Variant) Bindings.adapt(key, keyBinding, VARIANT_BINDING);
				TreeSet<Variant> ids = getIds();
				Variant ceilingId = ids.ceiling(nodeId);
				if (ceilingId == null) return null;				
				return Bindings.adapt(ceilingId, VARIANT_BINDING, keyBinding);
			} catch (AdaptException e) {
				throw new AccessorException( e );
			}
		}

		@Override
		public Object getHigherKey(Binding keyBinding, Object key)
				throws AccessorException {
			try {
				Variant nodeId = (Variant) Bindings.adapt(key, keyBinding, VARIANT_BINDING);
				TreeSet<Variant> ids = getIds();
				Variant higherId = ids.higher(nodeId);
				if (higherId == null) return null;				
				return Bindings.adapt(higherId, VARIANT_BINDING, keyBinding);
			} catch (AdaptException e) {
				throw new AccessorException( e );
			}
		}

		@Override
		public void apply(List<Event> cs, LinkedList<Event> rollback) throws AccessorException {
			try {
				boolean makeRollback = rollback != null;
				ArrayList<Event> single = new ArrayList<Event>();
				for (Event e : cs) {
					if (e.reference==null) {
						Event rbe = applyLocal(e, makeRollback);
						if (makeRollback) {
							rbe.reference = e.reference;
							rollback.addFirst( rbe );
						}					
					} else {
						Accessor sa = getAccessor(e.reference);
						// Apply changes
						single.clear();
						Event noRefEvent = e.clone(null);
						single.add(noRefEvent);
						sa.apply(single, rollback);
					}
				}
			} catch (AccessorConstructionException ae) {
				throw new AccessorException(ae);
			}				
		}		
		
		Event applyLocal(Event e, boolean makeRollback) throws AccessorException {
			Event rollback = null;
			
			if (e instanceof ValueAssigned) {
				ValueAssigned va = (ValueAssigned) e;
				if (makeRollback) {
					Binding binding = Bindings.getBinding(type());
					rollback = new ValueAssigned(binding, getValue(binding));				
				}
				setValue(va.newValue.getBinding(), va.newValue.getValue());
				return rollback;
			} else if (e instanceof MapEntryAdded) {
				MapEntryAdded ea = (MapEntryAdded) e;
				if (ea.key==null) throw new AccessorException("Cannot apply entry added event because key is missing");
				if (ea.value==null) throw new AccessorException("Cannot apply entry added event because value is missing");
				boolean hadValue = containsKey(ea.key.getBinding(), ea.key.getValue());
				if (hadValue) throw new AccessorException("Could not add entry to key that already existed");
				
				if (makeRollback) {				
					rollback = new MapEntryRemoved( ea.key );
				}
				
				put(ea.key.getBinding(), ea.key.getValue(), ea.value.getBinding(), ea.value.getValue());
				
			} else if (e instanceof MapEntryRemoved) {
				MapEntryRemoved er = (MapEntryRemoved) e;
				
				if (makeRollback) {
					boolean hadValue = containsKey(er.key.getBinding(), er.key.getValue());
					
					if (hadValue) {				
						Object oldValueObj = get(er.key.getBinding(), er.key.getValue(), NODE_BINDING);
						Variant oldKey = er.key;
						Variant oldValue = new Variant(NODE_BINDING, oldValueObj);
						rollback = new MapEntryAdded(oldKey, oldValue);
					} else {
						rollback = new MapEntryRemoved( er.key.clone() );
					}
				}
				
				remove( er.key.getBinding(), er.key.getValue() );
				
			} else throw new AccessorException("Cannot apply "+e.getClass().getName()+" to Map Type");
			
			return rollback;
		}
		
		
	}
	
}

