/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.serialization;

import gnu.trove.TObjectIntHashMap;

import java.io.IOException;
import java.nio.ByteOrder;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.List;

import org.simantics.databoard.Bindings;
import org.simantics.databoard.binding.ArrayBinding;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.BooleanBinding;
import org.simantics.databoard.binding.ByteBinding;
import org.simantics.databoard.binding.DoubleBinding;
import org.simantics.databoard.binding.FloatBinding;
import org.simantics.databoard.binding.IntegerBinding;
import org.simantics.databoard.binding.LongBinding;
import org.simantics.databoard.binding.MapBinding;
import org.simantics.databoard.binding.OptionalBinding;
import org.simantics.databoard.binding.RecordBinding;
import org.simantics.databoard.binding.StringBinding;
import org.simantics.databoard.binding.UnionBinding;
import org.simantics.databoard.binding.VariantBinding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.java.BooleanArrayBinding;
import org.simantics.databoard.binding.java.ByteArrayBinding;
import org.simantics.databoard.binding.java.DoubleArrayBinding;
import org.simantics.databoard.binding.java.FloatArrayBinding;
import org.simantics.databoard.binding.java.IntArrayBinding;
import org.simantics.databoard.binding.java.LongArrayBinding;
import org.simantics.databoard.serialization.Serializer.CompositeSerializer;
import org.simantics.databoard.serialization.Serializer.NonRecursiveSerializer;
import org.simantics.databoard.serialization.Serializer.RecursiveSerializer;
import org.simantics.databoard.type.ArrayType;
import org.simantics.databoard.type.BooleanType;
import org.simantics.databoard.type.ByteType;
import org.simantics.databoard.type.DataType;
import org.simantics.databoard.type.DoubleType;
import org.simantics.databoard.type.FloatType;
import org.simantics.databoard.type.IntegerType;
import org.simantics.databoard.type.LongType;
import org.simantics.databoard.type.RecordType;
import org.simantics.databoard.type.UnionType;
import org.simantics.databoard.util.Range;
import org.simantics.databoard.util.binary.BinaryReadable;
import org.simantics.databoard.util.binary.BinaryWriteable;

/**
 * BinarySerializationFormat serializes instances to byte format.
 * BinarySerializationFormat does not throw SerializationConstructionException.
 * 
 * The serialization notation is available at 
 * <a href="https://www.simantics.org/wiki/index.php/Org.simantics.datatype_Manual#Binary_Serialization">Binary Serialization format</a>
 * 
 * TODO Ensure java.lang.IndexOutOfBoundsException and other exceptions are not thrown when deserializing
 *      wrong binary data. Instead throw SerializationException
 */
public class BinarySerializationFormat implements SerializationFormat {

	public static final BinarySerializationFormat INSTANCE = new BinarySerializationFormat();  
	
	public static final Charset UTF8 = Charset.forName("utf-8");
	
	public static class ArraySerializer extends CompositeSerializer {

		Range length;
		Integer fixedLength;
		Integer fixedSize;
		Integer fixedSizeOfComponent;
		Serializer componentSerializer;
		ArrayBinding binding;
		
		public ArraySerializer(Serializer componentSerializer,
			ArrayBinding binding) {
			super( IsReferableQuery.isReferable( binding.type() ) != Result.No );
			ArrayType arrayType = (ArrayType) binding.type();
			this.componentSerializer = componentSerializer;
			this.binding = binding;
			this.length = arrayType.getLength();
			fixedSizeOfComponent = componentSerializer.getConstantSize();
			if (fixedSizeOfComponent!=null && length!=null && length.getLower().equals(length.getUpper()) && length.getLower().getValue()!=null)
			{
				fixedLength = length.getLower().getValue().intValue();
				fixedSize = fixedLength * fixedSizeOfComponent;
			}
		}

		@Override
		public Object deserialize(BinaryReadable in, List<Object> identities) throws IOException, SerializationException, BindingException {
			int length = fixedLength != null ? fixedLength : in.getInt();			
			ArrayList<Object> temp = new ArrayList<Object>(length);
			for(int i=0;i<length;++i)
				temp.add(componentSerializer.deserialize(in, identities));
			return binding.create(length, temp.iterator());
		}
		
		@Override
		public void skip(BinaryReadable in, List<Object> identities)
		throws IOException, SerializationException {
			if (fixedSize!=null) {
				in.skip(fixedSize);
				return;
			}
			int length = fixedLength != null ? fixedLength : in.getInt();
			
			if (fixedSizeOfComponent!=null) {
				// Skip all elements
				in.skip(length * fixedSizeOfComponent);
			} else {
				// Skip each element individualy
				for(int i=0;i<length;++i)
					componentSerializer.skip(in, identities);
			}
		}

		@Override
		public void serialize(BinaryWriteable out, TObjectIntHashMap<Object> identities, Object obj) throws IOException, BindingException, SerializationException {
			int length = binding.size(obj);
			if (fixedLength==null) {
				out.putInt(length);
			} else { 
				if (length!=fixedLength)
					throw new SerializationException("Unexpected array length "+length+", size is restricted to "+fixedLength);
			}
			
			// Serialize elements
			for(int i=0;i<length;++i)
				componentSerializer.serialize(out, identities, binding.get(obj, i));			
		}

		@Override
		public Integer getConstantSize() {
			return fixedSize;
		}

		@Override
		public int getSize(Object obj, TObjectIntHashMap<Object> identities) throws SerializationException, BindingException {
			if (fixedSize!=null) return fixedSize;
			int length = binding.size(obj);
			if (fixedSizeOfComponent!=null)
				return 4 + fixedSizeOfComponent * length;
			int result = 4;
			for(int i=0;i<length;++i)
				result += componentSerializer.getSize(binding.get(obj, i), identities);
			return result;
		}
		
	}
	
	public static class FloatArraySerializer extends NonRecursiveSerializer {

		Range length;
		Integer fixedLength, fixedSize;
		
		public FloatArraySerializer(ArrayBinding binding)
		{			
			ArrayType arrayType = (ArrayType) binding.type();
			this.length = arrayType.getLength();
			
			if (length!=null && length.getLower().equals(length.getUpper()) && length.getLower().getValue()!=null)
			{
				fixedLength = length.getLower().getValue().intValue();
				fixedSize = fixedLength * 4;
			}
		}
		
		@Override
		public Object deserialize(BinaryReadable in)
		throws IOException {
			int length = fixedLength != null ? fixedLength : in.getInt();			
			float[] array = new float[length];
			for(int i=0;i<array.length;++i)
				array[i] = in.getFloat();
			return array;
		}

		@Override
		public void skip(BinaryReadable in)
		throws IOException {
			int length = fixedSize != null ? fixedLength : in.getInt();			
			in.skip( length * 4 );
		}
		
		@Override
		public void serialize(BinaryWriteable out, Object obj)
			throws IOException {
			float[] array = (float[])obj;
			if (fixedSize==null) 
				out.putInt(array.length);
			for(float f : array)
				out.putFloat(f);
		}

		@Override
		public Integer getConstantSize() {
			return fixedSize;
		}

		@Override
		public int getSize(Object obj) throws SerializationException, BindingException {
			if (fixedSize!=null) return fixedSize;
			float[] array = (float[])obj;			
			return getSizeOfPutLengthForArray(array.length) + 4 * array.length;
		}
		
	}

	public static class DoubleArraySerializer extends NonRecursiveSerializer {

		Range length;
		Integer fixedLength, fixedSize;
		
		public DoubleArraySerializer(ArrayBinding binding)
		{
			ArrayType arrayType = (ArrayType) binding.type();			
			this.length = arrayType.getLength();
			if (length!=null && length.getLower().equals(length.getUpper()) && length.getLower().getValue()!=null)
			{
				fixedLength = length.getLower().getValue().intValue();
				fixedSize = fixedLength * 8;
			}
		}
		
		@Override
		public Object deserialize(BinaryReadable in)
		throws IOException 
		{
			int length = fixedSize != null ? fixedLength : in.getInt();			
			double[] array = new double[length];
			for(int i=0;i<array.length;++i)
				array[i] = in.getDouble();
			return array;
		}

		@Override
		public void skip(BinaryReadable in)
		throws IOException {
			int length = fixedSize != null ? fixedLength : in.getInt();			
			in.skip(length * 8);
		}
		
		@Override
		public void serialize(BinaryWriteable out, Object obj)
		throws IOException 
		{
			double[] array = (double[])obj;
			if (fixedSize==null) 
				out.putInt(array.length);
			for(double f : array)
				out.putDouble(f);
		}

		@Override
		public Integer getConstantSize() {
			return fixedSize;
		}

		@Override
		public int getSize(Object obj) throws SerializationException, BindingException {
			if (fixedSize!=null) return fixedSize;
			double[] array = (double[])obj;			
			return getSizeOfPutLengthForArray(array.length) + 8 * array.length;
		}
		
	}


	public static class LongArraySerializer extends NonRecursiveSerializer {

		Range length;
		Integer fixedLength, fixedSize;
		
		public LongArraySerializer(ArrayBinding binding)
		{
			ArrayType arrayType = (ArrayType) binding.type();
			this.length = arrayType.getLength();
			if (length!=null && length.getLower().equals(length.getUpper()) && length.getLower().getValue()!=null)
			{
				fixedLength = length.getLower().getValue().intValue();
				fixedSize = fixedLength * 8;
			}
		}
		
		@Override
		public Object deserialize(BinaryReadable in)
		throws IOException 
		{
			int length = fixedSize != null ? fixedLength : in.getInt();			
			long[] array = new long[length];
			for(int i=0;i<array.length;++i)
				array[i] = in.getLong();
			return array;
		}

		@Override
		public void skip(BinaryReadable in)
		throws IOException {
			int length = fixedSize != null ? fixedLength : in.getInt();			
			in.skip(8* length);			
		}
		
		@Override
		public void serialize(BinaryWriteable out, Object obj)
		throws IOException 
		{
			long[] array = (long[])obj;
			if (fixedSize==null) 
				out.putInt(array.length);
			for(long f : array)
				out.putLong(f);
		}

		@Override
		public Integer getConstantSize() {
			return fixedSize;
		}

		@Override
		public int getSize(Object obj) throws SerializationException, BindingException {
			if (fixedSize!=null) return fixedSize;
			long[] array = (long[])obj;			
			return getSizeOfPutLengthForArray(array.length) + 8 * array.length;
		}
		
	}

	
	public static class IntArraySerializer extends NonRecursiveSerializer {

		Range length;
		Integer fixedLength, fixedSize;
		
		public IntArraySerializer(ArrayBinding binding)
		{
			ArrayType arrayType = (ArrayType) binding.type();
			this.length = arrayType.getLength();
			if (length!=null && length.getLower().equals(length.getUpper()) && length.getLower().getValue()!=null)
			{
				fixedLength = length.getLower().getValue().intValue();
				fixedSize = fixedLength * 4;
			}
		}
		
		@Override
		public Object deserialize(BinaryReadable in)
		throws IOException 
		{
			int length = fixedSize != null ? fixedLength : in.getInt();			
			int[] array = new int[length];
			for(int i=0;i<array.length;++i)
				array[i] = in.getInt();
			return array;
		}

		@Override
		public void skip(BinaryReadable in)
		throws IOException {
			int length = fixedSize != null ? fixedLength : in.getInt();			
			in.skip(length*4);
		}
		
		@Override
		public void serialize(BinaryWriteable out, Object obj)
		throws IOException 
		{
			int[] array = (int[])obj;
			if (fixedSize==null) 
				out.putInt(array.length);
			for(int f : array)
				out.putInt(f);
		}

		@Override
		public Integer getConstantSize() {
			return fixedSize;
		}

		@Override
		public int getSize(Object obj) throws SerializationException, BindingException {
			if (fixedSize!=null) return fixedSize;
			int[] array = (int[])obj;			
			return getSizeOfPutLengthForArray(array.length) + 4 * array.length;
		}
		
	}


	public static class ByteArraySerializer extends NonRecursiveSerializer {

		Range length;
		Integer fixedLength, fixedSize;
		
		public ByteArraySerializer(ArrayBinding binding)
		{
			ArrayType arrayType = (ArrayType) binding.type();
			this.length = arrayType.getLength();
			if (length!=null && length.getLower().equals(length.getUpper()) && length.getLower().getValue()!=null)
			{
				fixedLength = length.getLower().getValue().intValue();
				fixedSize = fixedLength * 1;
			}
		}
		
		@Override
		public Object deserialize(BinaryReadable in)
		throws IOException 
		{
			int length = fixedSize != null ? fixedLength : in.getInt();			
			byte[] array = new byte[length];
			in.get(array);
			return array;
		}

		@Override
		public void skip(BinaryReadable in)
		throws IOException {
			int length = fixedSize != null ? fixedLength : in.getInt();			
			in.skip( length );
		}
		
		@Override
		public void serialize(BinaryWriteable out, Object obj)
		throws IOException 
		{
			byte[] array = (byte[])obj;
			if (fixedSize==null) 
				out.putInt(array.length);
			out.put(array);
		}

		@Override
		public Integer getConstantSize() {
			return fixedSize;
		}

		@Override
		public int getSize(Object obj) throws SerializationException, BindingException {
			if (fixedSize!=null) return fixedSize;
			byte[] array = (byte[])obj;			
			return getSizeOfPutLengthForArray(array.length) + array.length;
		}
		
	}
	
	public static class BooleanArraySerializer extends NonRecursiveSerializer {

		Range length;
		Integer fixedLength, fixedSize;
		
		public BooleanArraySerializer(ArrayBinding binding)
		{
			ArrayType arrayType = (ArrayType) binding.type();
			this.length = arrayType.getLength();
			if (length!=null && length.getLower().equals(length.getUpper()) && length.getLower().getValue()!=null)
			{
				fixedLength = length.getLower().getValue().intValue();
				fixedSize = fixedLength * 1;
			}
		}
		
		@Override
		public Object deserialize(BinaryReadable in)
		throws IOException, SerializationException 
		{
			int length = fixedLength != null ? fixedLength : in.getInt();			
			boolean[] array = new boolean[length];
			for(int i=0;i<array.length;++i) {
				byte value = in.get();
				if (value==0) array[i] = false;
				else if (value==1) array[i] = true;
				else throw new SerializationException("Unexpected value \""+value+"\" for boolean");
			}
			return array;
		}

		@Override
		public void skip(BinaryReadable in)
				throws IOException {
			int length = fixedSize != null ? fixedLength : in.getInt();			
			in.skip(length);
		}
		
		@Override
		public void serialize(BinaryWriteable out, Object obj)
		throws IOException 
		{
			boolean[] array = (boolean[])obj;
			if (fixedLength==null) 
				out.putInt(array.length);
			for(boolean f : array)
				out.put( (byte) (f ? 1 : 0) );
		}

		@Override
		public Integer getConstantSize() {
			return fixedSize;
		}

		@Override
		public int getSize(Object obj) throws SerializationException, BindingException {
			if (fixedSize!=null) return fixedSize;
			boolean[] array = (boolean[])obj;			
			return getSizeOfPutLengthForArray(array.length) + array.length;
		}
		
	}
	
	@Override
	public Serializer visit(ArrayBinding b) {
		ArrayType type = (ArrayType) b.type();
			
		if( b instanceof FloatArrayBinding && type.componentType instanceof FloatType )
			return new FloatArraySerializer(b);
		if(b instanceof DoubleArrayBinding && type.componentType instanceof DoubleType )
			return new DoubleArraySerializer(b);
		if(b instanceof IntArrayBinding && type.componentType instanceof IntegerType )
			return new IntArraySerializer(b);
		if(b instanceof ByteArrayBinding && type.componentType instanceof ByteType )
			return new ByteArraySerializer(b);
		if(b instanceof BooleanArrayBinding && type.componentType instanceof BooleanType )
			return new BooleanArraySerializer(b);
		if(b instanceof LongArrayBinding && type.componentType instanceof LongType )
			return new LongArraySerializer(b);
			
		Serializer componentSerializer = b.getComponentBinding().serializerUnchecked(this);
			
		return new ArraySerializer(componentSerializer, b);
	}
	
	public static class BooleanSerializer extends NonRecursiveSerializer {

		BooleanBinding binding;
		
		private BooleanSerializer(BooleanBinding binding) {
			this.binding = binding;
		}
		
		@Override
		public Object deserialize(BinaryReadable in) throws IOException, SerializationException, BindingException {
			boolean value = in.get() != 0; 
			return binding.create(value);
		}
		
		@Override
		public void skip(BinaryReadable in)
				throws IOException {
			in.skip(1);			
		}

		@Override
		public void serialize(BinaryWriteable out, Object obj)
			throws IOException, BindingException {
			boolean value = binding.getValue_(obj);
			out.put( (byte) (value ? 1 : 0) );
		}

		@Override
		public Integer getConstantSize() {
			return 1;
		}

		@Override
		public int getSize(Object obj) throws SerializationException, BindingException {
			return 1;
		}
		
	}

	@Override
	public Serializer visit(BooleanBinding b) {
		return new BooleanSerializer(b);
	}

	public static class DoubleSerializer extends NonRecursiveSerializer {

		DoubleBinding binding;
		
		private DoubleSerializer(DoubleBinding binding) {this.binding = binding;}
		
		@Override
		public Object deserialize(BinaryReadable in) throws IOException, SerializationException, BindingException {
			double value = in.getDouble();
			return binding.create(value);
		}

		@Override
		public void skip(BinaryReadable in)
				throws IOException {
			in.skip(8);			
		}
		
		@Override
		public void serialize(BinaryWriteable out, Object obj)
			throws IOException, BindingException {
			double value = binding.getValue_(obj);
			out.putDouble(value);			
		}

		@Override
		public Integer getConstantSize() {
			return 8;
		}

		@Override
		public int getSize(Object obj) throws SerializationException, BindingException {
			return 8;
		}
		
	}
	
	@Override
	public Serializer visit(DoubleBinding b) {
		return new DoubleSerializer(b);
	}
	
	public static class FloatSerializer extends NonRecursiveSerializer {
		
		FloatBinding binding;
		
		private FloatSerializer(FloatBinding binding) {this.binding = binding;}
		
		@Override
		public Object deserialize(BinaryReadable in) throws IOException, SerializationException, BindingException {
			float value = in.getFloat(); 
			return binding.create(value);
		}

		@Override
		public void skip(BinaryReadable in)
				throws IOException {
			in.skip(4);			
		}
		
		@Override
		public void serialize(BinaryWriteable out, Object obj)
			throws IOException, BindingException {
			float value = binding.getValue_(obj);
			out.putFloat(value);			
		}

		@Override
		public Integer getConstantSize() {
			return 4;
		}

		@Override
		public int getSize(Object obj) throws SerializationException, BindingException {
			return 4;
		}
		
	}	

	@Override
	public Serializer visit(FloatBinding b) {
		return new FloatSerializer(b);
	}
	
	public static class IntSerializer extends NonRecursiveSerializer {

		IntegerBinding binding;
		
		private IntSerializer(IntegerBinding binding) { this.binding = binding; }
		
		@Override
		public Object deserialize(BinaryReadable in) throws IOException, SerializationException, BindingException {
			int value = in.getInt();
			return binding.create(value);
		}
		
		@Override
		public void skip(BinaryReadable in)
				throws IOException {
			in.skip(4);			
		}

		@Override
		public void serialize(BinaryWriteable out, Object obj)
			throws IOException, BindingException {
			int value = binding.getValue_(obj);
			out.putInt(value);			
		}

		@Override
		public Integer getConstantSize() {
			return 4;
		}

		@Override
		public int getSize(Object obj) throws SerializationException, BindingException {
			return 4;
		}
		
	}	

	@Override
	public Serializer visit(IntegerBinding b) {
		return new IntSerializer(b);
	}

	
	public static class ByteSerializer extends NonRecursiveSerializer {

		ByteBinding binding;
		
		private ByteSerializer(ByteBinding binding) {
			this.binding = binding;
		}
		
		@Override
		public Object deserialize(BinaryReadable in) throws IOException, SerializationException, BindingException {
			byte value = in.get();
			return binding.create( value );
		}

		@Override
		public void skip(BinaryReadable in)
				throws IOException {
			in.skip(1);			
		}
		
		@Override
		public void serialize(BinaryWriteable out, Object obj)
			throws IOException, BindingException {
			byte value = binding.getValue_(obj);
			out.put(value);			
		}

		@Override
		public Integer getConstantSize() {
			return 1;
		}

		@Override
		public int getSize(Object obj) throws SerializationException, BindingException {
			return 1;
		}	
	}	
	
	@Override
	public Serializer visit(ByteBinding b) {
		return new ByteSerializer(b);
	}
	
	public static class LongSerializer extends NonRecursiveSerializer {

		LongBinding binding;
		
		private LongSerializer(LongBinding binding) {this.binding = binding;}
		
		@Override
		public Object deserialize(BinaryReadable in) throws IOException, SerializationException, BindingException {
			long value = in.getLong(); 
			return binding.create(value);
		}

		@Override
		public void skip(BinaryReadable in)
				throws IOException {
			in.skip(8);			
		}
		
		@Override
		public void serialize(BinaryWriteable out, Object obj)
			throws IOException, BindingException {
			long value = binding.getValue_(obj);
			out.putLong( value );
		}

		@Override
		public Integer getConstantSize() {
			return 8;
		}

		@Override
		public int getSize(Object obj) throws SerializationException, BindingException {
			return 8;
		}
		
	}	
	
	@Override
	public Serializer visit(LongBinding b) {
		return new LongSerializer(b);
	}
	
	public static class OptionalSerializer extends CompositeSerializer {

		OptionalBinding binding;
		Serializer componentSerializer;		
		Integer fixedSize;
		
		public OptionalSerializer(Serializer componentSerializer, OptionalBinding binding) {
			super( IsReferableQuery.isReferable( binding.type() ) != Result.No );
			this.componentSerializer = componentSerializer;
			this.binding = binding;
			Integer fixedSizeOfComponent = componentSerializer.getConstantSize();
			if (fixedSizeOfComponent!=null && fixedSizeOfComponent==0) {
				fixedSize = fixedSizeOfComponent + 1;
			}
		}

		@Override
		public Object deserialize(BinaryReadable in, List<Object> identities) throws IOException, SerializationException, BindingException {
			byte x = in.get();
			if (x == 0) return binding.createNoValue();
			else if (x == 1) {
				Object componentValue = componentSerializer.deserialize(in, identities);
				return binding.createValue(componentValue);
			}
			else throw new SerializationException("Unexpected marker for option "+x+" 0 or 1 expected.");
		}

		@Override
		public void skip(BinaryReadable in, List<Object> identities) throws IOException, SerializationException {
			int x = in.get();
			if (x==1) componentSerializer.skip(in, identities);
		}
		
		@Override
		public void serialize(BinaryWriteable out, TObjectIntHashMap<Object> identities, Object obj) throws IOException, SerializationException, BindingException {
			if (!binding.hasValue(obj)) {
				out.put((byte)0);
			} else {
				out.put((byte)1);
				Object componentValue = binding.getValue(obj);
				componentSerializer.serialize(out, identities, componentValue);
			}
		}

		@Override
		public Integer getConstantSize() {
			return fixedSize;
		}

		@Override
		public int getSize(Object obj, TObjectIntHashMap<Object> identities) throws SerializationException, BindingException {
			if (fixedSize!=null) return fixedSize;
			if (!binding.hasValue(obj)) return 1;
			Object componentValue = binding.getValue(obj);
			return 1 + componentSerializer.getSize(componentValue, identities);
		}
		
	}	

	@Override
	public Serializer visit(OptionalBinding b) {
		return new OptionalSerializer(b.getComponentBinding().serializerUnchecked(this), b);
	}
	
	public static class GenericRecordSerializer extends CompositeSerializer {

		RecordBinding binding;
		Serializer[] componentSerializers;
		Integer fixedSize;
		
		public GenericRecordSerializer(RecordBinding binding) {		
			super( IsReferableQuery.isReferable( binding.type() ) != Result.No );
			this.binding = binding;			
		}
		
		@Override
		public void postConstruction(SerializationFormat format) 
		throws RuntimeException 
		{
			Binding[] componentBindings = binding.getComponentBindings();
			componentSerializers = new Serializer[componentBindings.length];
			for(int i=0;i < componentSerializers.length; i++) 
				componentSerializers[i] = componentBindings[i].serializerUnchecked(format);
	
			fixedSize = null;
			for (Serializer componentSerializer : componentSerializers) {
				Integer componentFixedSize = componentSerializer.getConstantSize();
				if (componentFixedSize==null) {
					fixedSize = null;
					break;
				}
				fixedSize = fixedSize==null ? componentFixedSize : fixedSize+componentFixedSize;
			}	
		}		

		@Override
		public Object deserialize(BinaryReadable in, List<Object> identities) throws IOException, SerializationException, BindingException {
			Object[] temp = new Object[componentSerializers.length];
			for(int i=0;i<componentSerializers.length;++i)
				temp[i] = componentSerializers[i].deserialize(in, identities);
			return binding.create(temp);
		}

		@Override
		public void skip(BinaryReadable in, List<Object> identities) throws IOException, SerializationException {
			if (fixedSize != null) {
				in.skip(fixedSize);
			} else {
				for(int i=0;i<componentSerializers.length;++i)
					componentSerializers[i].skip(in, identities);				
			}
		}		
		
		@Override
		public void serialize(BinaryWriteable out, TObjectIntHashMap<Object> identities, Object obj) throws IOException, BindingException, SerializationException {
			for(int i=0;i<componentSerializers.length;++i)
				componentSerializers[i].serialize(out, identities, binding.getComponent(obj, i));
		}

		@Override
		public Integer getConstantSize() {			
			return fixedSize;
		}

		@Override
		public int getSize(Object obj, TObjectIntHashMap<Object> identities) throws SerializationException, BindingException {
			if (fixedSize!=null) return fixedSize;
			int result = 0;
			for(int i=0;i<componentSerializers.length;++i)
				result += componentSerializers[i].getSize( binding.getComponent(obj, i), identities );			
			return result;
		}
		
	}
	
	public static class ReferableRecordSerializer extends RecursiveSerializer {

		RecordBinding binding;
		Serializer[] componentSerializers;
		
		/** The combined size elements, or <code>null</code> */ 
		Integer fixedSizeOfContent;
		
		public ReferableRecordSerializer(RecordBinding binding) {
			this.binding = binding;
		}
		
		@Override
		public void postConstruction(SerializationFormat format)
		throws RuntimeException 
		{
			Binding[] componentBindings = binding.getComponentBindings();
			componentSerializers = new Serializer[componentBindings.length];
			for(int i=0;i < componentSerializers.length; i++) 
				componentSerializers[i] = componentBindings[i].serializerUnchecked(format);
				
			fixedSizeOfContent = null;
			for (Serializer componentSerializer : componentSerializers) {
				Integer componentFixedSize = componentSerializer.getConstantSize();
				if (componentFixedSize==null) {
					fixedSizeOfContent = null;
					break;
				}
				fixedSizeOfContent = fixedSizeOfContent==null ? componentFixedSize : fixedSizeOfContent + componentFixedSize; 
			}
		}

		boolean referable() {
			return ((RecordType)binding.type()).referable;
		}
		
		@Override
		public Object deserialize(BinaryReadable in, List<Object> identities) throws IOException, SerializationException, BindingException {
			int id = in.getInt();
			if(id > 0)
				return identities.get(id-1);			
			
			Object result = binding.createPartial();
			identities.add(result);
			
			Object[] temp = new Object[componentSerializers.length];
			for(int i=0;i<componentSerializers.length;++i)
				temp[i] = componentSerializers[i].deserialize(in, identities);
			binding.setComponents(result, temp);
			
			return result;
		}
		
		@Override
		public void skip(BinaryReadable in, List<Object> identities) throws IOException, SerializationException {
			
			int id = in.getInt();
			if(id > 0) return;			
			
			for(int i=0;i<componentSerializers.length;++i)
				componentSerializers[i].skip(in, identities);
		}

		@Override
		public void serialize(BinaryWriteable out, TObjectIntHashMap<Object> identities, Object obj) throws IOException, BindingException, SerializationException {
			int id = identities.get(obj);
			if(id > 0) {
				out.putInt(id);
			}
			else {
				out.putInt(0);
				identities.put(obj, identities.size() + 1);
				for(int i=0;i<componentSerializers.length;++i)
					componentSerializers[i].serialize(out, identities, binding.getComponent(obj, i));
			}
		}

		@Override
		public Integer getConstantSize() {
			return null;
		}

		@Override
		public int getSize(Object obj, TObjectIntHashMap<Object> identities) throws SerializationException, BindingException {
			int id = identities.get(obj);
			if (id>0) return 4;
			identities.put(obj, identities.size()+1);
			if (fixedSizeOfContent!=null) return 4+fixedSizeOfContent;
			int result = 4;			
			for(int i=0;i<componentSerializers.length;++i)
				result += componentSerializers[i].getSize( binding.getComponent(obj, i), identities );			
			return result;
		}
		
	}

	@Override
	public Serializer visit(RecordBinding b) {
		if(((RecordType)b.type()).referable)
			return new ReferableRecordSerializer(b);
		else
			return new GenericRecordSerializer(b);
	}

	public static class StringSerializer extends NonRecursiveSerializer {

		StringBinding binding;
		
		public StringSerializer(StringBinding binding) {
			this.binding = binding;
		}

		@Override
		public Object deserialize(BinaryReadable in) throws IOException, SerializationException, BindingException {
			int length = getLength(in);
			byte[] bytes = new byte[length];
			in.get(bytes);
			return binding.create(new String(bytes, UTF8));
		}

		@Override
		public void skip(BinaryReadable in) throws IOException, SerializationException {
			int length = getLength(in);
			in.skip(length);
		}
		
		@Override
		public void serialize(BinaryWriteable out, Object obj) throws IOException, BindingException, SerializationException {			
			byte[] bytes = binding.getValue(obj).getBytes(UTF8);
			putLength(out, bytes.length);	
			out.put(bytes);
		}

		@Override
		public Integer getConstantSize() {			
			return null;
		}

		@Override
		public int getSize(Object obj) throws SerializationException, BindingException {
			String string = binding.getValue(obj);
			int stringByteLength = getStringUTF8EncodedByteLength( string ); 
			int lengthLength = getSizeOfPutLength( stringByteLength ); 				
			return lengthLength + stringByteLength;
		}
		
	}
	
	@Override
	public Serializer visit(StringBinding b) {
		return new StringSerializer(b);
	}

	public static class GenericUnionSerializer extends CompositeSerializer {

		UnionBinding binding;
		Serializer[] componentSerializers;
		int tagTypeCount;
		Integer fixedSize;
		
		public GenericUnionSerializer(UnionBinding binding)				
		{
			super( IsReferableQuery.isReferable( binding.type() ) != Result.No );
			tagTypeCount = ((UnionType) binding.type()).components.length;
			this.binding = binding;
			
			// Check if all component serializers are the same size
			fixedSize = null;
		}
		
		@Override
		public void postConstruction(SerializationFormat format) 
		throws RuntimeException 
		{
			Binding[] componentBindings = binding.getComponentBindings();
			componentSerializers = new Serializer[componentBindings.length];
			for(int i=0;i < componentSerializers.length; i++) { 
				componentSerializers[i] = componentBindings[i].serializerUnchecked(format);
			}
				
			// Calculate Fixed Size
			for (Serializer s : componentSerializers)
			{
				Integer fixedSizeOfComponentSerializer = s.getConstantSize();
				if (fixedSizeOfComponentSerializer==null) {
					fixedSize = null;
					break;
				}
				if (fixedSize!=null && fixedSizeOfComponentSerializer!=fixedSize) {
					fixedSize = null;
					break;
				}
				fixedSize = fixedSizeOfComponentSerializer;
			}
			if (componentSerializers.length == 0) fixedSize = 0;
			if (fixedSize!=null) fixedSize += getUIntLength(tagTypeCount-1);
		}

		@Override
		public Object deserialize(BinaryReadable in, List<Object> identities) throws IOException, SerializationException, BindingException {
			int tag = getUInt(in, tagTypeCount-1);
			return binding.create(tag, componentSerializers[tag].deserialize(in, identities));
		}

		@Override
		public void skip(BinaryReadable in, List<Object> identities) throws IOException, SerializationException {
			int tag = getUInt(in, tagTypeCount-1);
			componentSerializers[tag].skip(in, identities);
		}
		
		@Override
		public void serialize(BinaryWriteable out, TObjectIntHashMap<Object> identities, Object obj) throws IOException, BindingException, SerializationException {
			int tag = binding.getTag(obj);
			putUInt(out, tag, tagTypeCount-1);
			componentSerializers[tag].serialize(out, identities, binding.getValue(obj));
		}

		@Override
		public Integer getConstantSize() {
			return fixedSize;
		}

		@Override
		public int getSize(Object obj, TObjectIntHashMap<Object> identities) throws SerializationException, BindingException {
			if (fixedSize!=null) return fixedSize;
			int tag = binding.getTag(obj);
			return getUIntLength( tagTypeCount-1 ) +
				componentSerializers[tag].getSize(binding.getValue(obj), identities);
		}
		
	}

	@Override
	public Serializer visit(UnionBinding b) {
		return new GenericUnionSerializer(b);
	}
	
	public static class VariantSerializer extends RecursiveSerializer {
		
		VariantBinding binding;
		Serializer datatypeSerializer;		
		
		public VariantSerializer(VariantBinding binding) {			
			this.binding = binding;			
		}
		
		@Override
		public void postConstruction(SerializationFormat format) 
		throws RuntimeException 
		{
			Binding dataTypeBinding = Bindings.getBindingUnchecked( DataType.class );
			this.datatypeSerializer = dataTypeBinding.serializerUnchecked( BinarySerializationFormat.INSTANCE );
		}		

		@Override
		public Object deserialize(BinaryReadable in, List<Object> identities) throws IOException, SerializationException, BindingException {			
			List<Object> typeIdentities = new ArrayList<Object>(1);				
			DataType type = (DataType) datatypeSerializer.deserialize(in, typeIdentities);
			Binding valueBinding = Bindings.getBinding(type);
			Serializer valueSerializer = (Serializer) valueBinding.serializerUnchecked(BinarySerializationFormat.INSTANCE);
			Object value = valueSerializer.deserialize(in, identities);
			return binding.create(value, valueBinding);
		}

		@Override
		public void skip(BinaryReadable in, List<Object> identities) throws IOException, SerializationException {
			try {
				List<Object> typeIdentities = new ArrayList<Object>(1);				
				DataType type = (DataType) datatypeSerializer.deserialize(in, typeIdentities);				
				Binding valueBinding = Bindings.getBinding(type);
				Serializer valueSerializer = valueBinding.serializerUnchecked(BinarySerializationFormat.INSTANCE);
				valueSerializer.skip(in, identities);
			} catch (BindingException e) {
				throw new SerializationException(e);
			}
		}		
		
		@Override
		public void serialize(BinaryWriteable out, TObjectIntHashMap<Object> identities, Object variant) 
		throws IOException, BindingException, SerializationException {
			TObjectIntHashMap<Object> typeIdentities = new TObjectIntHashMap<Object>(1);
			DataType type = binding.getType(variant);
			datatypeSerializer.serialize(out, typeIdentities, type);				
				
			Binding valueBinding = binding.getBinding(variant);
			Object value = binding.getValue(variant, valueBinding);
				
			Serializer valueSerializer = valueBinding.serializerUnchecked(BinarySerializationFormat.INSTANCE);
			valueSerializer.serialize(out, identities, value);
		}

		@Override
		public Integer getConstantSize() {			
			return null;
		}

		@Override
		public int getSize(Object variant, TObjectIntHashMap<Object> identities) 
		throws SerializationException, BindingException 
		{
			TObjectIntHashMap<Object> typeIdentities = new TObjectIntHashMap<Object>(1);
			DataType type = binding.getType(variant);
			int size = datatypeSerializer.getSize(type, typeIdentities);
				
			Binding valueBinding = binding.getBinding(variant);
			Object value = binding.getValue(variant, valueBinding);
				
			Serializer valueSerializer = valueBinding.serializerUnchecked(BinarySerializationFormat.INSTANCE);
			size += valueSerializer.getSize(value, identities);
			return size;
		}
		
	}
	
	@Override
	public Serializer visit(VariantBinding b) {
		return new VariantSerializer(b);
	}

	public static class MapSerializer extends CompositeSerializer {

		Integer fixedSizeOfKey, fixedSizeOfValue, fixedSizeOfEntry;
		Serializer keySerializer, valueSerializer;
		MapBinding binding;
		
		public MapSerializer(Serializer keySerializer, Serializer valueSerializer, MapBinding binding) {
			super( IsReferableQuery.isReferable( binding.type() ) != Result.No );
			this.keySerializer = keySerializer;
			this.valueSerializer = valueSerializer;
			this.binding = binding;
			
			fixedSizeOfKey = keySerializer.getConstantSize();
			fixedSizeOfValue = valueSerializer.getConstantSize();
			fixedSizeOfEntry = (fixedSizeOfKey!=null && fixedSizeOfValue!=null) ? (fixedSizeOfKey+fixedSizeOfValue) : null; 
		}

		@Override
		public Object deserialize(BinaryReadable in, List<Object> identities) throws IOException, SerializationException, BindingException {
			int length = in.getInt();
			Object keys[] = new Object[length];
			Object values[] = new Object[length];
			
			for(int i=0;i<length;++i) {
				keys[i] = keySerializer.deserialize(in, identities);
				values[i] = valueSerializer.deserialize(in, identities);				
			}
			return binding.create(keys, values);
		}
		
		@Override
		public void skip(BinaryReadable in, List<Object> identities)
		throws IOException, SerializationException {
			int length = in.getInt();
			if (fixedSizeOfEntry!=null) {
				in.skip( length * fixedSizeOfEntry );
				return;
			}
			for(int i=0;i<length;++i) {
				if (fixedSizeOfKey!=null) {
					in.skip(fixedSizeOfKey);
				} else {
					keySerializer.skip(in, identities);
				}
				
				if (fixedSizeOfValue!=null) {
					in.skip(fixedSizeOfValue);
				} else {
					valueSerializer.skip(in, identities);
				}
			}
		}

		@Override
		public void serialize(BinaryWriteable out, TObjectIntHashMap<Object> identities, Object map) throws IOException, BindingException, SerializationException {
			int length = binding.size(map);
			//if (binding.isOrderedMap(map)) 
			{
				Object[] keys = new Object[length];
				Object[] values = new Object[length];
				binding.getAll(map, keys, values);
				out.putInt(length);
				for(int i=0;i<length;++i) {
					Object key = keys[i];
					Object value = values[i];
					keySerializer.serialize(out, identities, key);
					valueSerializer.serialize(out, identities, value);				
				}
			}
			/*
			else {
				// Sort keys by putting into a treemap
				TreeMap<Object, Object> copyMap = new TreeMap<Object, Object>(binding.getKeyBinding());
				binding.getAll(map, copyMap);
				Iterator<Entry<Object, Object>> iter = copyMap.entrySet().iterator();
				putLength(out, length);
				while (iter.hasNext()) {
					Entry<Object, Object> e = iter.next();
					Object key = e.getKey();
					Object value = e.getValue();
					keySerializer.serialize(out, identities, key);
					valueSerializer.serialize(out, identities, value);				
				}
			}*/
		}

		@Override
		public Integer getConstantSize() {
			return null;
		}

		@Override
		public int getSize(Object obj, TObjectIntHashMap<Object> identities) throws SerializationException, BindingException {			
			int length = binding.size(obj);
			if (fixedSizeOfEntry!=null)
				return 4 + fixedSizeOfEntry * length;
			int result = 4;
			Object keys[] = binding.getKeys(obj);
			for(int i=0;i<length;++i) {
				Object key = keys[i];
				Object value = binding.get(obj, key);
				result += keySerializer.getSize(key, identities);
				result += valueSerializer.getSize(value, identities);
			}
			return result;
		}
		
	}	
	
	@Override
	public Serializer visit(MapBinding b) {
		return new MapSerializer(
				b.getKeyBinding().serializerUnchecked(this),
				b.getValueBinding().serializerUnchecked(this),
				b);
	}
	/**
	 * Get the number of bytes it takes to output length field of value <code>length</code>.
	 *  
	 * @param length length value
	 * @return bytes required
	 */
	public static int getSizeOfPutLength(int length)
	{
		if(length < 0x80) return 1;		
		if(length < 0x4080) return 2;
		if(length < 0x204000) return 3;
		if(length < 0x10200000) return 4;
		return 5;
	}
	
	public static int getSizeOfPutLengthForArray(int length) {
		return 4;
	}
	
	public static void putLength(BinaryWriteable out, int length) throws IOException {		
		if(length < 0x80) {
			out.put((byte)length);
		}
		else {
			length -= 0x80;
			if(length < 0x4000) {
				out.put((byte)((length&0x3f) | 0x80));
				out.put((byte)(length>>>6));
			}
			else {
				length -= 0x4000;
				if(length < 0x200000) {
					out.put((byte)((length&0x1f) | 0xc0));
					out.put((byte)((length>>>5)&0xff));
					out.put((byte)((length>>>13)&0xff));	
				}
				else {
					length -= 0x200000;
					if(length < 0x10000000) {
						out.put((byte)((length&0x0f) | 0xe0));
						out.put((byte)((length>>>4)&0xff));
						out.put((byte)((length>>>12)&0xff));	
						out.put((byte)((length>>>20)&0xff));
					}
					else {
						length -= 0x10000000;
						out.put((byte)((length&0x07) | 0xf0));
						out.put((byte)((length>>>3)&0xff));
						out.put((byte)((length>>>11)&0xff));	
						out.put((byte)((length>>>19)&0xff));
						out.put((byte)((length>>>27)&0xff));
					}
				}				
			}
		}	
	}

	
	public static int getLength(BinaryReadable in) throws IOException {
		int length = in.get()&0xff; 
		if(length >= 0x80) {
			if(length >= 0xc0) {
				if(length >= 0xe0) {
					if(length >= 0xf0) {
						length &= 0x0f;
						length += ((in.get()&0xff)<<3);
						length += ((in.get()&0xff)<<11);
						length += ((in.get()&0xff)<<19);
						length += 0x10204080;
					}
					else {
						length &= 0x1f;
						length += ((in.get()&0xff)<<4);
						length += ((in.get()&0xff)<<12);
						length += ((in.get()&0xff)<<20);
						length += 0x204080;
					}
				}
				else {
					length &= 0x3f;
					length += ((in.get()&0xff)<<5);
					length += ((in.get()&0xff)<<13);
					length += 0x4080;
				}
			}
			else {
				length &= 0x7f;
				length += ((in.get()&0xff)<<6);
				length += 0x80;
			}
		}
		return length;
	}
	
	public static void putUInt24(BinaryWriteable out, int value) throws IOException {
		if (out.order() == ByteOrder.BIG_ENDIAN) {
			out.put((byte)(value >> 16));
			out.put((byte)(value >> 8));
			out.put((byte)value);
		} else {
			out.put((byte)value);
			out.put((byte)(value >> 8));
			out.put((byte)(value >> 16));
		}
	}
	
	public static int getUInt24(BinaryReadable in) throws IOException {
		if (in.order() == ByteOrder.BIG_ENDIAN)
		{		
			return 
				( in.get() << 16) | 
				( in.get() << 8) |
				( in.get() );
		} else {
			return 
				( in.get() ) |
				( in.get() << 8) |
				( in.get() << 16); 
		}		
	}
	
	public static void putUInt(BinaryWriteable out, int value, int maxValue)
	throws IOException {
		if (maxValue==0) {}
		else if (maxValue<0x100) {
			out.put((byte) value);
		} else if (maxValue<0x10000) {
			out.putShort((short) value);
		} else if (maxValue<0x1000000) {
			putUInt24(out, value);
		} else {
			out.putInt(value);
		}
	}
	
	public static int getUInt(BinaryReadable in, int maxValue)
	throws IOException
	{
		if (maxValue==0) return 0;
		if (maxValue<0x100) {
			return in.get() & 0xFF;
		} else if (maxValue<0x10000) {
			return in.getShort() & 0xFFFF;
		} else if (maxValue<0x1000000) {
			return getUInt24(in) & 0xFFFFFF;
		} else {
			return in.getInt();
		}		
	}
	
	public static int getUIntLength(int maxValue)
	{
		if (maxValue==0) {
			return 0;
		} else if (maxValue<0x100) {
			return 1;
		} else if (maxValue<0x10000) {
			return 2;
		} else if (maxValue<0x1000000) {
			return 3;
		} else {
			return 4;
		}
	}
	
	public static int getStringUTF8EncodedByteLength(String string)
	{
		// TODO ?May not work properly properly on some characters that are encoded with 2 chars in UTF-16? 
		
		// Correct
		//return string.getBytes(UTF8).length;
		
		// http://en.wikipedia.org/wiki/UTF-8
		int result = 0;
		int length = string.length();
		for (int i=0; i<length; i++)
		{
			char c = string.charAt(i);
			if (c>=0 && c<=0x7f) {
				result += 1;
			} else if (c>=0x80 && c<=0x07ff) {
				result += 2;
			} else if (c>=0xD800 && c<=0xDFFF) {
				result += 1;
			} else if (c>=0x800 && c<=0xffff) {
				result += 3;
			}
			// Not really used as char is 16-bit
			else if (c>=0x10000 && c<=0x10ffff) {
				result += 4;
			} else if (c>=0x110000 && c<=0x1FFFFF) {
				result += 4;
			} else {
				// NOT IN RFC 3629
				result += 5;
			}
		}
		return result;				
	}
	
}
